import unittest
import os
import random

from datetime import date
from pathlib import Path
from unittest import skip

# from PCANBasic import *
from random import randint
import time

from mytoolit.old.network import Network
from mytoolit.old.MyToolItCommands import (
    ActiveState,
    BlueToothDeviceNr,
    BluetoothTime,
    byte_list_to_int,
    EepromPage,
    int_to_mac_address,
    MyToolItBlock,
    MyToolItEeprom,
    MyToolItProductData,
    MyToolItStatData,
    MyToolItStreaming,
    MyToolItSystem,
    NodeState,
    NetworkState,
    payload2Hex,
    sArray2String,
    SystemCommandBlueTooth,
)
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr
from mytoolit.old.MyToolItStu import TestConfig
from mytoolit.config import settings
from mytoolit.utility.environment import add_commander_path_to_environment

sVersion = TestConfig["Version"]
sLogFile = "TestStu.txt"
repo_root = str(Path(__file__).parent.parent.parent.parent)
sLogLocation = f"{repo_root}/"
sHomeLocation = "../../SimplicityStudio/v4_workspace/STU/"
sSilabsCommanderLocation = "../../SimplicityStudio/SimplicityCommander/"
sAdapterSerialNo = "440116697"
sBoardType = "BGM111A256V2"


class TestStu(unittest.TestCase):
    """
    This class is used for automated internal verification of the Stationary Transceiving Unit (STU)
    """

    def setUp(self):
        self.sHomeLocation = sHomeLocation
        self.sBuildLocation = sHomeLocation + "builds/" + sVersion
        self.sBootloader = (
            sHomeLocation + "builds/" + "BootloaderOtaBgm111.s37"
        )
        self.sAdapterSerialNo = sAdapterSerialNo
        self.sBoardType = sBoardType
        self.sSilabsCommander = "commander"
        self.bError = False
        self.Can = Network(
            sender=MyToolItNetworkNr["SPU1"],
            receiver=MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info("TestCase: " + str(self._testMethodName))
        self.vSilabsAdapterReset()
        if "test0000FirmwareFlash" != self._testMethodName:
            self.Can.CanTimeStampStart(self._resetStu()["CanTime"])
            self.sStuAddr = int_to_mac_address(
                self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"])
            )
            self.Can.logger.info("STU BlueTooth Address: " + self.sStuAddr)
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
            self._statusWords()
            self._StuWDog()
        self.Can.logger.info(
            "_______________________________________________________________________________________________________________"
        )
        self.Can.logger.info("Start")

    def tearDown(self):
        self.Can.logger.info("Fin")
        self.Can.logger.info(
            "_______________________________________________________________________________________________________________"
        )
        if False == self.Can.bError:
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
            if "test0000FirmwareFlash" == self._testMethodName:
                self.sStuAddr = hex(
                    self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"])
                )
                self.Can.logger.info("STU BlueTooth Address: " + self.sStuAddr)
                self._StuWDog()
            self._statusWords()
        self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("Test Time End Time Stamp")
        if False != self.Can.bError:
            self.bError = True
        self.Can.__exit__()

    def _test_has_failed(self):
        """
        Checks that a test case has failed or not
        """
        for _method, error in self._outcome.errors:
            if error:
                return True
        return False

    def _resetStu(self, retries=5, log=True):
        """
        Reset Stationary Transceiving Unit
        """
        self.Can.bConnected = False
        return self.Can.reset_node("STU1", retries=retries, log=log)

    def _StuWDog(self):
        """
        Retrieve Watchdog Counter of ST
        """
        WdogCounter = byte_list_to_int(
            self.Can.statisticalData(
                MyToolItNetworkNr["STU1"], MyToolItStatData["Wdog"]
            )[:4]
        )
        self.Can.logger.info("WatchDog Counter: " + str(WdogCounter))
        return WdogCounter

    def _statusWords(self):
        """
        Retrieve all status words
        """
        self.Can.logger.info(
            "STU Status Word: {}".format(
                self.Can.node_status(MyToolItNetworkNr["STU1"])
            )
        )
        self.Can.logger.info(
            "STU Error Word: {}".format(
                self.Can.error_status(MyToolItNetworkNr["STU1"])
            )
        )

    def vEepromWritePage(self, iPage, value):
        """
        Write Page by value
        """
        au8Content = [value] * 4
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            au8Payload = [iPage, 0xFF & offset, 4, 0] + au8Content
            self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Write"],
                au8Payload,
            )
        self.Can.logger.info(
            "Page Write Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )

    def vEepromReadPage(self, iPage, value):
        """
        Read page and check content
        """
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            au8Payload = [iPage, 0xFF & offset, 4, 0, 0, 0, 0, 0]
            index = self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Read"],
                au8Payload,
            )
            dataReadBack = self.Can.getReadMessageData(index)
            for dataByte in dataReadBack[4:]:
                self.assertEqual(dataByte, value)
        self.Can.logger.info(
            "Page Read Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )

    def vConnectSth1Dev0(self):
        """
        Connect to STH1 by device number 1
        """
        self.Can.logger.info("Connect")
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("Connect to Bluetooth Device")
        for _i in range(0, BluetoothTime["Connect"]):
            time.sleep(1)
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break

        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Check to be connected")
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            time.sleep(1)

    def vSilabsAdapterReset(self):
        """
        Reset the Silicon Labs Adapter
        """
        self.Can.logger.info("Reset Adapter " + self.sAdapterSerialNo)
        sSystemCall = self.sSilabsCommander + " adapter reset "
        sSystemCall += "--serialno " + self.sAdapterSerialNo
        sSystemCall += ">>" + sLogLocation + "AdapterReset.txt"
        if os.name == "nt":
            sSystemCall = sSystemCall.replace("/", "\\")
            os.system(sSystemCall)
        # for mac and linux(here, os.name is 'posix')
        else:
            os.system(sSystemCall)
        time.sleep(4)

    @skip("Already covered by test-stu")
    def test0000FirmwareFlash(self):
        """Upload firmware"""

        try:
            os.remove(sLogLocation + "ManufacturingCreateResport.txt")
        except:
            pass
        try:
            os.remove(sLogLocation + "ManufacturingFlashResport.txt")
        except:
            pass
        sSystemCall = self.sSilabsCommander + " convert "
        sSystemCall += self.sBootloader + " "
        sSystemCall += self.sBuildLocation + "/firmware_client.s37 "
        sSystemCall += "--patch 0x0fe04000:0x00 --patch 0x0fe041F8:0xFD "
        sSystemCall += (
            "-o "
            + self.sBuildLocation
            + "/manufacturingImageStu"
            + sVersion
            + ".hex "
        )
        sSystemCall += "-d " + self.sBoardType + " "
        sSystemCall += ">> " + sLogLocation
        sSystemCall += "ManufacturingCreateResport.txt"
        if os.name == "nt":
            sSystemCall = sSystemCall.replace("/", "\\")
            os.system(sSystemCall)
        # for mac and linux(here, os.name is 'posix')
        else:
            os.system(sSystemCall)
        tFile = open(
            sLogLocation + "ManufacturingCreateResport.txt",
            "r",
            encoding="utf-8",
        )
        asData = tFile.readlines()
        tFile.close()
        self.assertEqual("DONE\n", asData[-1])
        sSystemCall = self.sSilabsCommander + " flash "
        sSystemCall += (
            self.sBuildLocation + "/manufacturingImageStu" + sVersion + ".hex "
        )
        sSystemCall += "--address 0x0 "
        sSystemCall += "--serialno " + self.sAdapterSerialNo + " "
        sSystemCall += "-d " + self.sBoardType + " "
        sSystemCall += ">> " + sLogLocation
        sSystemCall += "ManufacturingFlashResport.txt"
        if os.name == "nt":
            sSystemCall = sSystemCall.replace("/", "\\")
            os.system(sSystemCall)
        # for mac and linux(here, os.name is 'posix')
        else:
            os.system(sSystemCall)
        tFile = open(
            sLogLocation + "ManufacturingFlashResport.txt",
            "r",
            encoding="utf-8",
        )
        asData = tFile.readlines()
        tFile.close()
        self.assertEqual(
            "range 0x0FE04000 - 0x0FE047FF (2 KB)\n", asData[-2][10:]
        )
        self.assertEqual("DONE\n", asData[-1])
        time.sleep(4)

    @skip("OTA update using `ota-dfu` is **very** unreliable")
    def test0001OverTheAirUpdate(self):
        """
        Test the over the air update
        """
        bCreate = os.path.isfile(self.sBuildLocation + "/OtaServer.gbl")
        bCreate = bCreate and os.path.isfile(
            self.sBuildLocation + "/OtaApploader.gbl"
        )
        bCreate = bCreate and os.path.isfile(
            self.sBuildLocation + "/OtaApploaderServer.gbl"
        )
        bCreate = not bCreate
        if False != bCreate:
            iRuns = 4
            iRuns += 1
            try:
                for i in range(1, iRuns):
                    os.remove(sLogLocation + str(i) + ".txt")
            except:
                pass

            try:
                os.remove(sLogLocation + "CreateReportOta.txt")
            except:
                pass
            try:
                os.remove(self.sBuildLocation + "/OtaClient.gbl")
            except:
                pass
            try:
                os.remove(self.sBuildLocation + "/OtaApploader.gbl")
            except:
                pass
            try:
                os.remove(self.sBuildLocation + "/OtaApploaderClient.gbl")
            except:
                pass

            self._resetStu()
            time.sleep(1)
            sSystemCall = (
                self.sHomeLocation + "/firmware_client/create_bl_files.bat "
            )
            sSystemCall += " -> " + sLogLocation
            sSystemCall += "CreateReportOta.txt"
            if os.name == "nt":
                sSystemCall = sSystemCall.replace("/", "\\")
                os.system(sSystemCall)
            # for mac and linux(here, os.name is 'posix')
            else:
                os.system(sSystemCall)
            os.rename(
                self.sHomeLocation
                + "/firmware_client/output_gbl/application.gbl",
                self.sBuildLocation + "/OtaClient.gbl",
            )
            os.rename(
                self.sHomeLocation
                + "/firmware_client/output_gbl/apploader.gbl",
                self.sBuildLocation + "/OtaApploader.gbl",
            )
            os.rename(
                self.sHomeLocation + "/firmware_client/output_gbl/full.gbl",
                self.sBuildLocation + "/OtaApploaderClient.gbl",
            )
        for i in range(1, iRuns):
            sSystemCall = self.sBuildLocation + "/ota-dfu.exe COM6 115200 "
            sSystemCall += self.sBuildLocation + "/OtaClient.gbl "
            sSystemCall += self.sStuAddr + " -> " + sLogLocation
            sSystemCall += "Ota" + str(i) + ".txt"
            if os.name == "nt":
                sSystemCall = sSystemCall.replace("/", "\\")
                os.system(sSystemCall)
            # for mac and linux(here, os.name is 'posix')
            else:
                os.system(sSystemCall)
            tFile = open(
                sLogLocation + "Ota" + str(i) + ".txt", "r", encoding="utf-8"
            )
            asData = tFile.readlines()
            tFile.close()
            self.assertEqual("Finishing DFU block...OK\n", asData[-2])
            self.assertEqual("Closing connection...OK\n", asData[-1])

    def test0005Ack(self):
        """
        Test Acknowledgement from STU  (⏱ 9 seconds)

        Write message and check identifier to be ack (No bError)
        """
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.Can.logger.info("Write Message")
        self.Can.WriteFrame(msg)
        self.Can.logger.info("Wait 200ms")
        time.sleep(0.25)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msgAckExpected = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["SPU1"], [0]
        )
        self.Can.logger.info(
            "Send ID: "
            + hex(msg.ID)
            + "; Expected ID: "
            + hex(msgAckExpected.ID)
            + "; Received ID: "
            + hex(self.Can.getReadMessage(-1).ID)
        )
        expectedData = ActiveState()
        expectedData.asbyte = 0
        expectedData.b.u2NodeState = NodeState["Application"]
        expectedData.b.u3NetworkState = NetworkState["Operating"]
        self.Can.logger.info(
            "Send Data: "
            + hex(0)
            + "; Expected Data: "
            + hex(expectedData.asbyte)
            + "; Received Data: "
            + hex(self.Can.getReadMessage(-1).DATA[0])
        )
        self.assertEqual(
            hex(msgAckExpected.ID), hex(self.Can.getReadMessage(-1).ID)
        )
        self.assertEqual(
            expectedData.asbyte, self.Can.getReadMessage(-1).DATA[0]
        )

    def test0006FirmwareVersion(self):
        """Check firmware version (⏱ 8 seconds)"""

        iIndex = self.Can.cmdSend(
            MyToolItNetworkNr["STU1"],
            MyToolItBlock["Product Data"],
            MyToolItProductData["Firmware Version"],
            [],
        )
        au8Version = self.Can.getReadMessageData(iIndex)
        au8Version = au8Version[-3:]
        sVersionRead = (
            "v"
            + str(au8Version[0])
            + "."
            + str(au8Version[1])
            + "."
            + str(au8Version[2])
        )
        self.Can.logger.info("Version: " + sVersionRead)
        self.assertEqual(sVersion, sVersionRead)

    def test0052MultiSend(self):
        """Send Multiple Frames without waiting for an ACK (⏱ 60 seconds)

        do ACK after 100 times send flooding to check functionality
        """
        self.Can.logger.info(
            "Send command 100 times, check number of write/reads and do ack"
            " test at the end; do that for 100 times"
        )
        for i in range(1, 101):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            for _j in range(1, 101):
                if 1 == randint(0, 1):
                    cmd = self.Can.CanCmd(
                        MyToolItBlock["System"],
                        MyToolItSystem["Get/Set State"],
                        1,
                        0,
                    )
                    message = self.Can.CanMessage20(
                        cmd,
                        MyToolItNetworkNr["SPU1"],
                        MyToolItNetworkNr["STU1"],
                        [0],
                    )
                else:
                    cmd = self.Can.CanCmd(
                        MyToolItBlock["System"],
                        MyToolItSystem["Bluetooth"],
                        1,
                        0,
                    )
                    message = self.Can.CanMessage20(
                        cmd,
                        MyToolItNetworkNr["SPU1"],
                        MyToolItNetworkNr["STU1"],
                        [
                            SystemCommandBlueTooth["DeviceCheckConnected"],
                            0,
                            0,
                            0,
                            0,
                            0,
                            0,
                            0,
                        ],
                    )
                self.Can.WriteFrame(message)
            time.sleep(0.5)
            self.Can.tWriteFrameWaitAckRetries(message, retries=0)

    def test0053MultiSendAck(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 13 seconds)

        Send->Ack->Send->Ack
        """

        self.Can.logger.info(
            "Send and get ACK for 1000 times AND do it with two messages"
            " randomly "
        )
        for i in range(1, 1001):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            if 1 == randint(0, 1):
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Get/Set State"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STU1"],
                    [0],
                )
            else:
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"], MyToolItSystem["Bluetooth"], 1, 0
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STU1"],
                    [
                        SystemCommandBlueTooth["DeviceCheckConnected"],
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                    ],
                )
            self.assertNotEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.test0005Ack()  # Test that it still works

    def test0054MultiSendMultiAckRetries(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 13 seconds)

        Send->Ack->Send->Ack, this also do a retry, tests the test framework - Multiple Messages
        """
        self.Can.logger.info(
            "Send and get ACK for 1000 times AND do it with two messages"
            " randomly "
        )
        for i in range(1, 1001):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            if 1 == randint(0, 1):
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Get/Set State"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STU1"],
                    [0],
                )
            else:
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"], MyToolItSystem["Bluetooth"], 1, 0
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STU1"],
                    [
                        SystemCommandBlueTooth["DeviceCheckConnected"],
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                    ],
                )
            self.assertNotEqual(
                "bError", self.Can.tWriteFrameWaitAckRetries(msg, retries=3)
            )
        self.test0005Ack()  # Test that it still works

    def test0055MultiSendSingleAckRetries(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 60 seconds)

        Send->Ack->Send->Ack, this also do a retry, tests the test framework - Single Message
        """
        self.Can.logger.info(
            "Send and get ACK for 1000 times AND do it with two messages"
            " randomly "
        )
        for _i in range(1, 10001):
            self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["System"],
                MyToolItSystem["Bluetooth"],
                [
                    SystemCommandBlueTooth["DeviceCheckConnected"],
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ],
                retries=0,
            )
        self.test0005Ack()  # Test that it still works

    def test0056SenderReceiver(self):
        """
        Send addressing same sender and receiver (⏱ 9 seconds)
        """
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        # Test that it still works
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0)

    def test0057ChristmasTree(self):
        """
        "Christmas Tree" packages (⏱ 14 seconds)
        """
        self.Can.logger.info("Error Request Frame from STU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Request Frame from SPU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Ack Frame from STU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Ack Frame from SPU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Ack Frame from STU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Ack Frame from SPU1 to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])

        # Test that it still works
        self.Can.logger.info("Normal Request to STU1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], [0]
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0)

    def test0101BlueToothConncectDeviceNr(self):
        """
        Connect and disconnect device (⏱ 7 minutes)

        check device number after each connect/disconnect to check correctness
        """
        self.Can.logger.info(
            "Connect and get Device Number, disconnect and get device number"
        )
        for i in range(0, 300):
            self.Can.logger.info("Loop Run: " + str(i))
            self.Can.logger.info("Connect")
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
            for _j in range(0, BluetoothTime["DeviceConnect"]):
                deviceNumbers = self.Can.iBlueToothConnectTotalScannedDeviceNr(
                    MyToolItNetworkNr["STU1"]
                )
                if 0 < deviceNumbers:
                    break
                time.sleep(1)
            self.Can.logger.info(
                "Number of available devices: " + str(deviceNumbers)
            )
            self.assertGreater(deviceNumbers, 0)
            if False != self.Can.bBlueToothDisconnect(
                MyToolItNetworkNr["STU1"]
            ):
                self.Can.logger.error("Bluetooth STH connected")
                self.assertEqual(False, True)
            deviceNumbers = self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            )
            self.Can.logger.info(
                "Number of available devices: " + str(deviceNumbers)
            )
            self.assertEqual(deviceNumbers, 0)

    def test0102BlueToothConnectDisconnectDevice(self):
        """
        Connect and disconnect to device 50 times (⏱ 3 minutes)

        The old version of this test would try to connect and disconnect
        500 times. Since this meant the test would fail most of the time we
        decreased the number of connection attempts to 50.
        """
        self.Can.logger.info(
            "Bluetooth connect command and check connected command and"
            " disconnect command"
        )
        totalConnectDisconnectTime = 0
        totalRuns = 50
        for i in range(0, totalRuns):
            startTime = time.time() * 1000
            self.Can.logger.info("Loop Run: " + str(i))
            self.vConnectSth1Dev0()
            if False != self.Can.bConnected:
                self.Can.logger.info("Bluetooth STH connected")
            else:
                self.Can.logger.error("Bluetooth STH not Connected")
            self.assertNotEqual(False, self.Can.bConnected)
            self.Can.logger.info("Disconnect from Bluetooth Device")
            if False != self.Can.bBlueToothDisconnect(
                MyToolItNetworkNr["STU1"]
            ):
                self.Can.logger.error("Bluetooth STH connected")
                self.assertEqual(False, True)
            else:
                self.Can.logger.info("Bluetooth STH not Connected")
            ConnectDissconnectTime = time.time() * 1000 - startTime
            self.Can.logger.info(
                "Time to connect and disconnect: "
                + str(ConnectDissconnectTime)
                + "ms"
            )
            totalConnectDisconnectTime += ConnectDissconnectTime
            time.sleep(1)
        totalConnectDisconnectTime /= totalRuns
        self.Can.logger.info(
            "Average Time to connect and disconnect: "
            + str(totalConnectDisconnectTime)
            + "ms"
        )

    def test0103BlueToothName(self):
        """
        Write name and get name (bluetooth command) (⏱ 3 minutes)
        """
        self.Can.logger.info("Bluetooth name command to STH")
        for _i in range(0, 10):
            self.Can.logger.info("Loop Run: " + str(_i))
            self.vConnectSth1Dev0()
            time.sleep(1)
            self.Can.logger.info("Write Walther0")
            self.Can.vBlueToothNameWrite(
                MyToolItNetworkNr["STH1"], 0, "Walther0"
            )
            self.Can.logger.info("Check Walther0")
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
            time.sleep(1)
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
            time.sleep(2)
            Name = self.Can.BlueToothNameGet(MyToolItNetworkNr["STU1"], 0)
            self.Can.logger.info("Received Name: " + Name)
            self.assertEqual("Walther0", Name)
            self.vConnectSth1Dev0()
            time.sleep(1)
            self.Can.logger.info("Write " + settings.sth.name)
            self.Can.vBlueToothNameWrite(
                MyToolItNetworkNr["STH1"], 0, settings.sth.name
            )
            self.Can.logger.info("Check " + settings.sth.name)
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
            time.sleep(2)
            time.sleep(BluetoothTime["Disconnect"])
            Name = self.Can.BlueToothNameGet(MyToolItNetworkNr["STU1"], 0)
            self.Can.logger.info("Received Name: " + Name)
            self.assertEqual(settings.sth.name, Name)
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
            time.sleep(2)

    def test0104BluetoothAddressDevices(self):
        """
        Check that Bluetooth addresses are listed correctly (⏱ 20 seconds)
        """
        for _i in range(0, 10):
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
            endTime = time.time() + 5
            while 0 == self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            ):
                if time.time() > endTime:
                    break
            time.sleep(0.5)
            devNrs = self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            )
            for devNr in range(0, devNrs):
                self.Can.logger.info("Device Number " + str(devNr))
                Address = self.Can.BlueToothAddressGet(
                    MyToolItNetworkNr["STU1"], devNr
                )
            self.Can.logger.info("Address: " + hex(Address))
            self.assertGreater(Address, 0)
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])

    def test0105BluetoothRssi(self):
        """
        Check that correct Bluetooth RSSIs are listed (⏱ 20 seconds)
        """
        for _i in range(0, 10):
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
            endTime = time.time() + 5
            while 0 == self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            ):
                if time.time() > endTime:
                    break
            time.sleep(0.5)
            Rssi = self.Can.BlueToothRssiGet(MyToolItNetworkNr["STU1"], 0)
            self.Can.logger.info("RSSI: " + str(int(Rssi)))
            self.assertNotEqual(Rssi, 127)
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])

    def test0106BluetoothRssiChange(self):
        """
        Check that correct Bluetooth RSSIs change (⏱ 11 seconds)
        """
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        endTime = time.time() + 5
        while 0 == self.Can.iBlueToothConnectTotalScannedDeviceNr(
            MyToolItNetworkNr["STU1"]
        ):
            if time.time() > endTime:
                break
        time.sleep(0.5)
        Rssi = []
        for _i in range(0, 500):
            Rssi.append(
                self.Can.BlueToothRssiGet(MyToolItNetworkNr["STU1"], 0)
            )
            self.Can.logger.info("RSSI: " + str(int(Rssi[-1])))
            self.assertNotEqual(Rssi, 127)
        self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        Rssi.sort()
        self.assertNotEqual(Rssi[1], Rssi[-2])

    def test0107BluetoothNameAddressRssi(self):
        """
        Check Bluetooth name, addresses and RSSIs (⏱ 9 seconds)
        """
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        endTime = time.time() + 5
        while 0 == self.Can.iBlueToothConnectTotalScannedDeviceNr(
            MyToolItNetworkNr["STU1"]
        ):
            if time.time() > endTime:
                break
        Name = self.Can.BlueToothNameGet(MyToolItNetworkNr["STU1"], 0)
        Address = self.Can.BlueToothAddressGet(MyToolItNetworkNr["STU1"], 0)
        Rssi = self.Can.BlueToothRssiGet(MyToolItNetworkNr["STU1"], 0)
        self.Can.logger.info("Name: " + Name)
        self.Can.logger.info("Address: " + hex(Address))
        self.Can.logger.info("RSSI: " + str(Rssi))
        self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])

    def test0110BlueToothConnectDisconnectDevicePolling(self):
        """
        Connect and disconnect to device 100 times (⏱ 3 minutes)

        Do it without time out, use connection check
        """
        self.Can.logger.info(
            "Bluetooth connect command and check connected command and"
            " disconnect command"
        )
        startTime = time.time() * 1000
        for _i in range(0, 100):
            self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"], 0)
            while 0 == self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            ):
                pass
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        endTime = time.time() * 1000
        ConnectDisconnectTime = endTime - startTime
        ConnectDisconnectTime /= 100
        self.Can.logger.info(
            "Average Time for connect and disconnect: "
            + str(ConnectDisconnectTime)
            + "ms"
        )

    def test0111BlueToothAddress(self):
        """
        Get Bluetooth Address (⏱ 8 seconds)
        """
        self.Can.logger.info("Get Bluetooth Address")
        self.Can.logger.info(
            "BlueTooth Address: "
            + hex(self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"]))
        )

    def test0112BlueToothConnectMac(self):
        """
        Connect to STH via Bluetooth Address (⏱ 13 seconds)
        """
        iRssiLimit = -80
        self.Can.logger.info("Connect via bluetooth address")
        self.Can.logger.info("Connect command")
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("Sleep 4s such that device list is up2date")
        time.sleep(5)
        iRssi = -100
        iRetries = 100
        self.Can.logger.info(
            "Rssi of device must be better than: " + str(iRssiLimit)
        )
        iDevPicked = -1
        iAddressReadback = -2
        while iRssiLimit > iRssi and 0 < iRetries:
            iRetries -= 1
            devNrs = self.Can.iBlueToothConnectTotalScannedDeviceNr(
                MyToolItNetworkNr["STU1"]
            )
            self.Can.logger.info("Devises Found: " + str(devNrs))
            iDevPicked = randint(0, devNrs - 1)
            self.Can.logger.info("Try to pick device: " + str(iDevPicked))
            iRssi = self.Can.BlueToothRssiGet(
                MyToolItNetworkNr["STU1"], iDevPicked
            )
            self.Can.logger.info("Rssi: " + str(iRssi))
        if 0 < iRetries:
            iAddress = self.Can.BlueToothAddressGet(
                MyToolItNetworkNr["STU1"], iDevPicked
            )
            iAddressReadback = self.Can.iBlueToothConnect2MacAddr(
                MyToolItNetworkNr["STU1"], iAddress
            )
        self.Can.logger.info("Try to get Address: " + hex(iAddress))
        self.Can.logger.info("Taken Address: " + hex(iAddressReadback))
        self.assertEqual(iAddressReadback, iAddress)
        self.Can.bBlueToothCheckConnect(MyToolItNetworkNr["STU1"])
        self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])

    def test0113DeviceNameChangeSTU(self):
        """
        Change Device Name of STU (⏱ 40 seconds)
        """
        self.Can.logger.info("Bluetooth name command to STU")
        for _i in range(0, 10):
            self.Can.logger.info("Loop Run: " + str(_i))
            self.Can.logger.info("Write Walther0")
            self.Can.vBlueToothNameWrite(
                MyToolItNetworkNr["STU1"],
                BlueToothDeviceNr["Self"],
                "Walther0",
            )
            self.Can.logger.info("Check Walther0")
            Name = self.Can.BlueToothNameGet(
                MyToolItNetworkNr["STU1"], BlueToothDeviceNr["Self"]
            )
            self.Can.logger.info("Received Name: " + Name)
            self.assertEqual("Walther0", Name)
            self.Can.logger.info("Write " + TestConfig["StuName"])
            self.Can.vBlueToothNameWrite(
                MyToolItNetworkNr["STU1"],
                BlueToothDeviceNr["Self"],
                TestConfig["StuName"],
            )
            self.Can.logger.info("Check " + TestConfig["StuName"])
            self._resetStu()
            Name = self.Can.BlueToothNameGet(
                MyToolItNetworkNr["STU1"], BlueToothDeviceNr["Self"]
            )
            self.Can.logger.info("Received Name: " + Name)
            self.assertEqual(TestConfig["StuName"], Name)
            time.sleep(1)

    def test0201MyToolItTestNotConnectedAck(self):
        """
        Send Message to STH without connecting (⏱ 10 seconds)

        Assumed result = not receiving anything. This especially tests the routing functionality.
        """
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.logger.info("Write Message")
        lastIndex = self.Can.GetReadArrayIndex()
        self.Can.WriteFrame(msg)
        self.Can.logger.info("Wait 2000ms")
        time.sleep(2)
        self.assertEqual(self.Can.GetReadArrayIndex(), lastIndex)

    def test0202MyToolItTestAck(self):
        """
        Send Message to STH with connecting (⏱ 16 seconds)

        Assumed result = receive correct ack. This especially tests the routing functionality.
        """
        expectedData = ActiveState()
        expectedData.asbyte = 0
        expectedData.b.u2NodeState = NodeState["Application"]
        expectedData.b.u3NetworkState = NetworkState["Operating"]
        self.vConnectSth1Dev0()
        time.sleep(3)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.logger.info("Write Message")
        self.Can.WriteFrame(msg)
        waitTime = 1
        self.Can.logger.info("Wait " + str(waitTime) + "s")
        time.sleep(waitTime)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msgAckExpected = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["STH1"],
            MyToolItNetworkNr["SPU1"],
            [expectedData.asbyte],
        )
        self.Can.logger.info(
            "Send ID: "
            + hex(msg.ID)
            + "; Expected ID: "
            + hex(msgAckExpected.ID)
            + "; Received ID: "
            + hex(self.Can.getReadMessage(-1).ID)
        )
        self.Can.logger.info(
            "Send Data: "
            + hex(0)
            + "; Expected Data: "
            + hex(expectedData.asbyte)
            + "; Received Data: "
            + hex(self.Can.getReadMessage(-1).DATA[0])
        )
        self.assertEqual(
            hex(msgAckExpected.ID), hex(self.Can.getReadMessage(-1).ID)
        )
        self.assertEqual(
            hex(msgAckExpected.DATA[0]),
            hex(self.Can.getReadMessage(-1).DATA[0]),
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0)
        self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])

    def test0203MyToolItTestWrongReceiver(self):
        """
        Send Message to STH with connecting (⏱ 45 seconds)

        Assumed result = receive correct ack. This especially tests the routing functionality.
        """
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        for i in range(MyToolItNetworkNr["STH2"], 32):
            if MyToolItNetworkNr["STU1"] != i:
                msg = self.Can.CanMessage20(
                    cmd, MyToolItNetworkNr["SPU1"], i, [0]
                )
                ack = self.Can.tWriteFrameWaitAck(msg)
                self.assertEqual("Error", ack[0])

        # Test that it still works
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0)

    def test0204RoutingMultiSend(self):
        """Send Multiple Frames without waiting for an ACK (⏱ 45 seconds)

        do ACK after 100 times send flooding to check functionality
        """
        self.Can.logger.info(
            "Send command 100 times over STU to STH, check number of"
            " write/reads and do ack test at the end; do that for 1000 times"
        )
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        for i in range(1, 101):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            for _j in range(1, 101):
                if 1 == randint(0, 1):
                    cmd = self.Can.CanCmd(
                        MyToolItBlock["System"],
                        MyToolItSystem["Get/Set State"],
                        1,
                        0,
                    )
                    message = self.Can.CanMessage20(
                        cmd,
                        MyToolItNetworkNr["SPU1"],
                        MyToolItNetworkNr["STH1"],
                        [0],
                    )
                else:
                    cmd = self.Can.CanCmd(
                        MyToolItBlock["System"],
                        MyToolItSystem["Node Status"],
                        1,
                        0,
                    )
                    message = self.Can.CanMessage20(
                        cmd,
                        MyToolItNetworkNr["SPU1"],
                        MyToolItNetworkNr["STH1"],
                        [0, 0, 0, 0, 0, 0, 0, 0],
                    )
                self.Can.WriteFrame(message)
            self.Can.reset()
            time.sleep(0.25)
            self.Can.tWriteFrameWaitAckRetries(message, retries=0)

    def test0205RoutingMultiSendAck(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 30 seconds)

        with routing: Send->Ack->Send->Ack
        """
        self.Can.logger.info(
            "Send and get ACK for 10000 times AND do it with two messages"
            " randomly "
        )
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        for i in range(1, 1001):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            if 1 == randint(0, 1):
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Get/Set State"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STH1"],
                    [0],
                )
            else:
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Node Status"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STH1"],
                    [0, 0, 0, 0, 0, 0, 0, 0],
                )
            self.assertNotEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])

    def test0206RoutingMultiSendAckRetries(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 30 seconds)

        Send->Ack->Send->Ack with routing, this also do a retry, tests the test framework - Multiple Messages
        """
        self.Can.logger.info(
            "Send and get ACK for 10000 times AND do it with two messages"
            " randomly "
        )
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        for i in range(1, 1001):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            if 1 == randint(0, 1):
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Get/Set State"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STH1"],
                    [0],
                )
            else:
                cmd = self.Can.CanCmd(
                    MyToolItBlock["System"],
                    MyToolItSystem["Node Status"],
                    1,
                    0,
                )
                msg = self.Can.CanMessage20(
                    cmd,
                    MyToolItNetworkNr["SPU1"],
                    MyToolItNetworkNr["STH1"],
                    [0, 0, 0, 0, 0, 0, 0, 0],
                )
            self.assertNotEqual(
                "bError", self.Can.tWriteFrameWaitAckRetries(msg, retries=0)
            )

    def test0207RoutingMultiSendSingleAckRetries(self):
        """Send Multiple Frames with waiting for an ACK (⏱ 30 seconds)

        Send->Ack->Send->Ack with routing, this also do a retry, tests the test framework - Single Message
        """

        self.Can.logger.info(
            "Send and get ACK for 10000 times AND do it with two messages"
            " randomly "
        )
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        for i in range(1, 1001):
            self.Can.logger.info(
                "Received Index: " + str(self.Can.GetReadArrayIndex())
            )
            self.Can.logger.info("Run: " + str(i))
            cmd = self.Can.CanCmd(
                MyToolItBlock["System"], MyToolItSystem["Node Status"], 1, 0
            )
            msg = self.Can.CanMessage20(
                cmd,
                MyToolItNetworkNr["SPU1"],
                MyToolItNetworkNr["STH1"],
                [0, 0, 0, 0, 0, 0, 0, 0],
            )
            self.assertNotEqual(
                "bError", self.Can.tWriteFrameWaitAckRetries(msg, retries=0)
            )

    def test0208RoutingSenderReceiver(self):
        """
        Send addressing same sender and receiver via Routing (⏱ 14 seconds)
        """
        self.Can.logger.info(
            "Connect to STH and send message with STH1=sender/receiver"
        )
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STH1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        # Test that it still works
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0, bErrorExit=False)

    def test0209RoutingChristmasTree(self):
        """
        "Christmas Tree" packages via routing (⏱ 20 seconds)
        """
        self.Can.logger.info("Error Request Frame from STH1 to STH1")
        self.Can.vBlueToothConnectConnect(MyToolItNetworkNr["STU1"])
        for _i in range(0, BluetoothTime["Connect"]):
            self.Can.logger.info("Device Connect to number 0")
            if False != self.Can.bBlueToothConnectDeviceConnect(
                MyToolItNetworkNr["STU1"], 0
            ):
                break
            time.sleep(1)
        for _i in range(0, BluetoothTime["Connect"]):
            if False != self.Can.bBlueToothCheckConnect(
                MyToolItNetworkNr["STU1"]
            ):
                break
            else:
                time.sleep(1)
        time.sleep(3)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STH1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Request Frame from SPU1 to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Ack Frame from STH1 to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STH1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Error Ack Frame from SPU1 to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 1
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Ack Frame from STH1 to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STH1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])
        self.Can.logger.info("Ack Frame from SPU1 to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.assertEqual("Error", self.Can.tWriteFrameWaitAck(msg)[0])

        # Test that it still works
        self.Can.logger.info("Normal Request to STH1")
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.tWriteFrameWaitAckRetries(msg, retries=0)

    def test0700StatisticsPowerOnCounterPowerOffCounter(self):
        """
        Check Power On and Power Off Counters  (⏱ 10 seconds)
        """
        PowerOnOff1 = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["PocPof"]
        )
        PowerOn1 = byte_list_to_int(PowerOnOff1[:4])
        PowerOff1 = byte_list_to_int(PowerOnOff1[4:])
        self._resetStu()
        PowerOnOff2 = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["PocPof"]
        )
        PowerOn2 = byte_list_to_int(PowerOnOff2[:4])
        PowerOff2 = byte_list_to_int(PowerOnOff2[4:])
        self.Can.logger.info(
            "PowerOnOff Payload before STU Reset: " + payload2Hex(PowerOnOff1)
        )
        self.Can.logger.info(
            "Power On Counter before STU Reset: " + str(PowerOn1)
        )
        self.Can.logger.info(
            "Power Off Counter before STU Reset: " + str(PowerOff1)
        )
        self.Can.logger.info(
            "PowerOnOff Payload after STU Reset: " + payload2Hex(PowerOnOff2)
        )
        self.Can.logger.info(
            "Power On Counter after STU Reset: " + str(PowerOn2)
        )
        self.Can.logger.info(
            "Power Off Counter after STU Reset: " + str(PowerOff2)
        )
        self.assertEqual(PowerOn1 + 1, PowerOn2)
        self.assertEqual(PowerOff1, PowerOff2)

    def test0701StatisticsOperatingSeconds(self):
        """
        Check Operating Seconds (⏱ 33 minutes)
        """
        u32EepromWriteRequestCounterTestStart = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        OperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
        )
        SecondsReset1 = byte_list_to_int(OperatingSeconds[:4])
        SecondsOveral1 = byte_list_to_int(OperatingSeconds[4:])
        time.sleep(60)
        OperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
        )
        SecondsReset2 = byte_list_to_int(OperatingSeconds[:4])
        SecondsOveral2 = byte_list_to_int(OperatingSeconds[4:])
        self._resetStu()
        OperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
        )
        SecondsReset3 = byte_list_to_int(OperatingSeconds[:4])
        SecondsOveral3 = byte_list_to_int(OperatingSeconds[4:])
        time.sleep(60 * 31)
        OperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
        )
        SecondsReset4 = byte_list_to_int(OperatingSeconds[:4])
        SecondsOveral4 = byte_list_to_int(OperatingSeconds[4:])
        self.Can.logger.info(
            "Operating Seconds since Reset: " + str(SecondsReset1)
        )
        self.Can.logger.info(
            "Operating Seconds since frist PowerOn: " + str(SecondsOveral1)
        )
        self.Can.logger.info(
            "Operating Seconds since Reset(+1 minute): " + str(SecondsReset2)
        )
        self.Can.logger.info(
            "Operating Seconds since frist PowerOn(+1minute): "
            + str(SecondsOveral2)
        )
        self.Can.logger.info(
            "Operating Seconds since Reset(After Disconnect/Connect): "
            + str(SecondsReset3)
        )
        self.Can.logger.info(
            "Operating Seconds since frist PowerOn(After Disconnect/Connect): "
            + str(SecondsOveral3)
        )
        self.Can.logger.info(
            "Operating Seconds since Reset(+30 minutes): " + str(SecondsReset4)
        )
        self.Can.logger.info(
            "Operating Seconds since frist PowerOn(+30minutes): "
            + str(SecondsOveral4)
        )
        u32EepromWriteRequestCounterTestEnd = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        u32EepromWriteRequsts = (
            u32EepromWriteRequestCounterTestEnd
            - u32EepromWriteRequestCounterTestStart
        )
        self.Can.logger.info(
            "EEPROM Write Requests during tests: " + str(u32EepromWriteRequsts)
        )
        self.assertEqual(
            1, u32EepromWriteRequsts
        )  # +1 due to operating seconds
        self.assertLess(SecondsReset1, 10)
        self.assertGreater(SecondsReset2, 60)
        self.assertLess(SecondsReset2, 70)
        self.assertLess(SecondsReset3, 10)
        self.assertGreater(SecondsReset4, 60 * 31 - 20)
        self.assertLess(SecondsReset4, 20 + 60 * 31)
        self.assertEqual(SecondsOveral1, SecondsOveral2)
        self.assertLess(SecondsOveral1 + 58, SecondsOveral3)
        self.assertGreater(SecondsOveral1 + 63, SecondsOveral3)
        self.assertLess(SecondsOveral1 + 58, SecondsOveral3)
        self.assertGreater(SecondsOveral1 + 63, SecondsOveral3)
        self.assertLess(SecondsOveral3 + 30 * 60 - 3, SecondsOveral4)
        self.assertGreater(SecondsOveral3 + 30 * 60 + 4, SecondsOveral4)

    def test0702WdogNotIncrementing(self):
        """
        Check Watchdog counter to not increment (⏱ 12 seconds)
        """
        WDogCounter1 = self._StuWDog()
        self._resetStu()
        WDogCounter2 = self._StuWDog()
        self._resetStu()
        self.Can.logger.info("Watchdog Counter at start: " + str(WDogCounter1))
        self.Can.logger.info(
            "Watchdog Counter after reset: " + str(WDogCounter2)
        )
        self.assertEqual(WDogCounter1, WDogCounter2)

    def test0703ProductionDate(self):
        """
        Check ProductionDate (⏱ 8 seconds)
        """
        sProductionDate = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["ProductionDate"]
        )
        sProductionDate = sArray2String(sProductionDate)
        year = int(sProductionDate[0:4])
        month = int(sProductionDate[4:6])
        day = int(sProductionDate[6:8])
        production_date = date(year, month, day)
        self.Can.logger.info(f"Production Date: {production_date}")
        self.assertEqual(settings.stu.production_date, production_date)

    def test0750StatisticPageWriteReadDeteministic(self):
        """
        Check EEPROM Read/Write - Deterministic data (⏱ 4 minutes)
        """
        uLoopRuns = 25
        time.sleep(2)
        u32EepromWriteRequestCounterTestStart = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        self.Can.logger.info("Save up EEPROM content")
        startData = []
        for offset in range(0, 256, 4):
            index = self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Read"],
                [EepromPage["Statistics"], 0xFF & offset, 4, 0, 0, 0, 0, 0],
            )
            dataReadBack = self.Can.getReadMessageData(index)
            startData.extend(dataReadBack[4:])

        # Test it self
        for _i in range(0, uLoopRuns):
            self.Can.logger.info("Next Run 12 Writes and Reads")
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
            self.vEepromWritePage(EepromPage["Statistics"], 0xAA)
            self.vEepromReadPage(EepromPage["Statistics"], 0xAA)
            self.vEepromWritePage(EepromPage["Statistics"], 0xFF)
            self.vEepromReadPage(EepromPage["Statistics"], 0xFF)
            self.vEepromWritePage(EepromPage["Statistics"], 0xAA)
            self.vEepromReadPage(EepromPage["Statistics"], 0xAA)
            self.vEepromWritePage(EepromPage["Statistics"], 0x00)
            self.vEepromReadPage(EepromPage["Statistics"], 0x00)
            self.vEepromWritePage(EepromPage["Statistics"], 0xAA)
            self.vEepromReadPage(EepromPage["Statistics"], 0xAA)
            self.vEepromWritePage(EepromPage["Statistics"], 0x55)
            self.vEepromReadPage(EepromPage["Statistics"], 0x55)
            self.vEepromWritePage(EepromPage["Statistics"], 0xAA)
            self.vEepromReadPage(EepromPage["Statistics"], 0xAA)
            self.vEepromWritePage(EepromPage["Statistics"], 0x55)
            self.vEepromReadPage(EepromPage["Statistics"], 0x55)
            self.vEepromWritePage(EepromPage["Statistics"], 0x00)
            self.vEepromReadPage(EepromPage["Statistics"], 0x00)
            self.vEepromWritePage(EepromPage["Statistics"], 0x55)
            self.vEepromReadPage(EepromPage["Statistics"], 0x55)
            self.vEepromWritePage(EepromPage["Statistics"], 0xFF)
            self.vEepromReadPage(EepromPage["Statistics"], 0xFF)
            self.vEepromWritePage(EepromPage["Statistics"], 0x55)
            self.vEepromReadPage(EepromPage["Statistics"], 0x55)

        # Write Back Page
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            payload = [EepromPage["Statistics"], 0xFF & offset, 4, 0]
            payload.extend(startData[offset : offset + 4])
            self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Write"],
                payload,
            )
        self.Can.logger.info(
            "Page Write Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )
        u32EepromWriteRequestCounterTestEnd = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        u32EepromWriteRequsts = (
            u32EepromWriteRequestCounterTestEnd
            - u32EepromWriteRequestCounterTestStart
        )
        self.Can.logger.info(
            "EEPROM Write Requests during tests: " + str(u32EepromWriteRequsts)
        )
        self.assertEqual(
            u32EepromWriteRequestCounterTestStart + 1,
            u32EepromWriteRequestCounterTestEnd,
        )  # +1 due to incrementing at first write

    def test0751StatisticPageWriteReadRandom(self):
        """
        Check EEPROM Read/Write - Deterministic data (⏱ 75 seconds)
        """
        uLoopRuns = 100
        u32EepromWriteRequestCounterTestStart = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        self.Can.logger.info("Save up EEPROM content")
        startData = []
        for offset in range(0, 256, 4):
            index = self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Read"],
                [EepromPage["Product Data"], 0xFF & offset, 4, 0, 0, 0, 0, 0],
            )
            dataReadBack = self.Can.getReadMessageData(index)
            startData.extend(dataReadBack[4:])

        # Test it self
        for _i in range(0, uLoopRuns):
            self.Can.logger.info("Next random Writes and Reads")
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
            au8ReadCheck = []
            for offset in range(0, 256, 4):
                au8Content = []
                for _j in range(0, 4):
                    u8Byte = int(random.random() * 0xFF)
                    au8Content.append(u8Byte)
                au8ReadCheck.extend(au8Content)
                au8Payload = [
                    EepromPage["Product Data"],
                    0xFF & offset,
                    4,
                    0,
                ] + au8Content
                self.Can.cmdSend(
                    MyToolItNetworkNr["STU1"],
                    MyToolItBlock["EEPROM"],
                    MyToolItEeprom["Write"],
                    au8Payload,
                )
            for offset in range(0, 256, 4):
                au8Payload = [
                    EepromPage["Product Data"],
                    0xFF & offset,
                    4,
                    0,
                    0,
                    0,
                    0,
                    0,
                ]
                index = self.Can.cmdSend(
                    MyToolItNetworkNr["STU1"],
                    MyToolItBlock["EEPROM"],
                    MyToolItEeprom["Read"],
                    au8Payload,
                )
                dataReadBack = self.Can.getReadMessageData(index)
                self.assertEqual(
                    dataReadBack[4:], au8ReadCheck[offset : offset + 4]
                )
            self.Can.logger.info("Fin random Writes and Reads")
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])

        # Write Back Page
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            payload = [EepromPage["Product Data"], 0xFF & offset, 4, 0]
            payload.extend(startData[offset : offset + 4])
            self.Can.cmdSend(
                MyToolItNetworkNr["STU1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Write"],
                payload,
            )
        self.Can.logger.info(
            "Page Write Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )
        self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        u32EepromWriteRequestCounterTestEnd = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        u32EepromWriteRequsts = (
            u32EepromWriteRequestCounterTestEnd
            - u32EepromWriteRequestCounterTestStart
        )
        self.Can.logger.info(
            "EEPROM Write Requests during tests: " + str(u32EepromWriteRequsts)
        )
        self.assertEqual(
            u32EepromWriteRequestCounterTestStart + 1,
            u32EepromWriteRequestCounterTestEnd,
        )  # +1 due to incrementing at first write

    def test0753EepromWriteRequestCounterPageSwitches(self):
        """
        Check that page switched do not yield to Writing EEPROM (⏱ 16 seconds)
        """
        time.sleep(1)
        uLoopRuns = 5
        u32EepromWriteRequestCounterTestStart = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        for _i in range(0, uLoopRuns):
            for sPage in EepromPage:
                self.Can.logger.info("Next Page")
                for offset in range(0, 256, 4):
                    au8Payload = [
                        EepromPage[sPage],
                        0xFF & offset,
                        4,
                        0,
                        0,
                        0,
                        0,
                        0,
                    ]
                    self.Can.cmdSend(
                        MyToolItNetworkNr["STU1"],
                        MyToolItBlock["EEPROM"],
                        MyToolItEeprom["Read"],
                        au8Payload,
                    )
        u32EepromWriteRequestCounterTestEnd = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        u32EepromWriteRequsts = (
            u32EepromWriteRequestCounterTestEnd
            - u32EepromWriteRequestCounterTestStart
        )
        self.Can.logger.info(
            "EEPROM Write Requests during tests: " + str(u32EepromWriteRequsts)
        )
        self.assertEqual(
            u32EepromWriteRequestCounterTestStart,
            u32EepromWriteRequestCounterTestEnd,
        )  # +1 due to incrementing at first write

    def test0754EepromWriteRequestCounterPageWriteSwitches(self):
        """
        Check that page switched with previews writes (⏱ 12 seconds)

        yield into to Writing EEPROM with the correct number of wirtes
        """
        time.sleep(1)
        uLoopRuns = 5
        uPageStart = 10
        uPageRuns = 6
        u32EepromWriteRequestCounterTestStart = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        for _i in range(0, uLoopRuns):
            for uPageOffset in range(0, uPageRuns):
                self.Can.logger.info("Next Page")
                self.Can.u32EepromWriteRequestCounter(
                    MyToolItNetworkNr["STU1"]
                )
                uPage = uPageOffset + uPageStart
                au8Payload = [uPage, 12, 4, 0, 0, 0, 0, 0]
                self.Can.cmdSend(
                    MyToolItNetworkNr["STU1"],
                    MyToolItBlock["EEPROM"],
                    MyToolItEeprom["Write"],
                    au8Payload,
                    log=False,
                )
                uPage = uPageOffset + 2
                uPage %= uPageRuns
                uPage += uPageStart
                au8Payload = [uPage, 12, 4, 0, 0, 0, 0, 0]
                self.Can.cmdSend(
                    MyToolItNetworkNr["STU1"],
                    MyToolItBlock["EEPROM"],
                    MyToolItEeprom["Read"],
                    au8Payload,
                    log=False,
                )
        u32EepromWriteRequestCounterTestEnd = (
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STU1"])
        )
        u32EepromWriteRequsts = (
            u32EepromWriteRequestCounterTestEnd
            - u32EepromWriteRequestCounterTestStart
        )
        self.Can.logger.info(
            "EEPROM Write Requests during tests: " + str(u32EepromWriteRequsts)
        )
        self.assertEqual(uPageRuns * uLoopRuns, u32EepromWriteRequsts)

    def test0900ErrorCmdVerbotenStu1(self):
        """
        Test that nothing happens when sending 0x0000 to STU1 (⏱ 24 seconds)
        """
        cmd = self.Can.CanCmd(0, 0, 1, 0)
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)
        cmd = self.Can.CanCmd(0, 0, 1, 1)
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)
        cmd = self.Can.CanCmd(0, 0, 0, 0)
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)
        cmd = self.Can.CanCmd(0, 0, 0, 1)
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)

    def test0901ErrorRequestErrorStu1(self):
        """
        Test that nothing happens when sending Request(1)/Error(1) (⏱ 16 seconds)
        """
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Reset"], 1, 1
        )
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)
        cmd = self.Can.CanCmd(
            MyToolItBlock["Streaming"], MyToolItStreaming["Data"], 1, 1
        )
        message = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STU1"], []
        )
        msgAck = self.Can.tWriteFrameWaitAckRetries(
            message, waitMs=1000, retries=3, bErrorExit=False
        )
        self.assertEqual("Error", msgAck)


def main():
    add_commander_path_to_environment()
    unittest.main(module=__name__, verbosity=2)


if __name__ == "__main__":
    main()



from asyncio import run
from time import time

from mytoolit.can import Message, Network


async def pfeifferl(
    identifier,
    channel_number=0,
    signal_type="cw",
    wait_time_seconds=0,
    duration_rf_seconds=0,
    power=0,
    antenna=1,
    packet_length=37,
    myprint=print,
):
    """exclusively broadcast RF test signals (cw/pn9),
    inherently prevents BLE/MYT operations during execution,
    STH is disconnected and advertising afterwards

    Byte_nr, arg : descr (type)
    - *,   identifier: STH-name (str)
    - 1,   channel_number: silabs BLE CH nr, 0-39 (uint8)
    - 2,   signal_type: cw/pn9 (str)
    - 3,   wait_time_seconds (till rf is turned on) (uint8)
    - 4-5, duration_rf_seconds (how long is RF turned on) (int)
    - 6,   RF power 0-32, default = 0 (uint8)
    - 7,   physical antenna, default = 1 (uint8)
    - 8,   length of RF packet, default = 37 (uint8)
    - *,   myprint redirector, default = print (fct_ptr)

    returns: execution_time
    """
    # # map signal_type to byte id # #
    if signal_type == "pn9":
        signal_type = 253
    elif signal_type == "cw":
        signal_type = 254
    else:
        raise Exception("signal type {} unknown - typo?".format(signal_type))

    # # open MYT network, send, return  # #
    async with Network() as network:
        node = "STH 1"
        start_time = time()

        await network.connect_sensor_device(identifier)
        name = await network.get_name(node)
        mac_address = await network.get_mac_address(node)
        myprint(
            f"Connected to sensor device “{name}” with MAC "
            f"address “{mac_address}”"
        )

        data = [
            channel_number,
            signal_type,
            wait_time_seconds,
            *list((duration_rf_seconds).to_bytes(2, byteorder="big")),
            power,
            antenna,
            packet_length,
        ]
        message = Message(
            block="Test",
            block_command=0x69,
            sender="SPU 1",
            receiver="STH 1",
            request=True,
            data=data,
        )
        myprint(f"Send message: {message}")
        answer = await network._request(message, "“Pfeifferl” test command")

        myprint(Message(answer))

        exec_time = time() - start_time
        myprint("\nExecution took {:.3} seconds".format(exec_time))

        return exec_time


if __name__ == "__main__":
    # call async fct via asyncio.run
    run(pfeifferl(identifier="Aladdin"))



# -- Imports ------------------------------------------------------------------

from argparse import Namespace
from asyncio import run, sleep
from functools import partial
from logging import basicConfig, getLogger
from pathlib import Path
from sys import stderr
from time import time
from typing import List

from can.interfaces.pcan import PcanError
from tqdm import tqdm

from mytoolit.can import Network
from mytoolit.can.adc import ADCConfiguration
from mytoolit.can.network import STHDeviceInfo, NetworkError
from mytoolit.cmdline.parse import parse_arguments
from mytoolit.config import ConfigurationUtility
from mytoolit.measurement import convert_raw_to_g, Storage


# -- Functions ----------------------------------------------------------------


def config(arguments: Namespace) -> None:
    """Open configuration file"""

    ConfigurationUtility.open_user_config()


async def dataloss(arguments: Namespace) -> None:
    """Check data loss at different sample rates"""

    identifier = arguments.identifier
    logger = getLogger(__name__)

    async with Network() as network:
        logger.info(f"Connecting to “{identifier}”")
        await network.connect_sensor_device(identifier)
        logger.info(f"Connected to “{identifier}”")

        sensor_range = await network.read_acceleration_sensor_range_in_g()
        conversion_to_g = partial(convert_raw_to_g, max_value=sensor_range)

        measurement_time_s = 10

        for oversampling_rate in (2**exponent for exponent in range(6, 10)):
            logger.info(f"Oversampling rate: {oversampling_rate}")
            config = ADCConfiguration(
                prescaler=2,
                acquisition_time=8,
                oversampling_rate=oversampling_rate,
            )
            await network.write_adc_configuration(**config)
            sample_rate = config.sample_rate()
            logger.info(f"Sample rate: {sample_rate} Hz")

            filepath = Path(f"Measurement {sample_rate} Hz.hdf5")
            with Storage(filepath.resolve()) as storage:
                progress = tqdm(
                    total=int(sample_rate * measurement_time_s),
                    desc="Read sensor data",
                    unit=" values",
                    leave=False,
                    disable=None,
                )

                start_time = time()
                try:
                    async with network.open_data_stream(first=True) as stream:
                        async for data in stream:
                            data.apply(conversion_to_g)
                            storage.add_streaming_data(data)
                            progress.update(3)
                            if time() - start_time >= measurement_time_s:
                                break
                except PcanError as error:
                    print(
                        f"Unable to collect streaming data: {error}",
                        file=stderr,
                    )

                storage.add_acceleration_meta(
                    "Sensor_Range", f"± {sensor_range/2} g₀"
                )

                progress.close()
            print(f"Stored measurement data in “{filepath}”")


async def list_sensor_devices(arguments: Namespace) -> None:
    """Print a list of available sensor devices

    Parameters
    ----------

    arguments:
        The given command line arguments

    """

    async with Network() as network:
        timeout = time() + 5
        sensor_devices: List[STHDeviceInfo] = []
        sensor_devices_before: List[STHDeviceInfo] = []

        # - First request for sensor devices will produce empty list
        # - Subsequent retries should provide all available sensor devices
        # - We wait until the number of sensor devices is larger than 1 and
        #   has not changed between one iteration or the timeout is reached
        while (
            len(sensor_devices) <= 0
            and time() < timeout
            or len(sensor_devices) != len(sensor_devices_before)
        ):
            sensor_devices_before = list(sensor_devices)
            sensor_devices = await network.get_sensor_devices()
            await sleep(0.5)

        for device in sensor_devices:
            print(device)


async def measure(arguments: Namespace) -> None:
    """Open measurement stream and store data

    Parameters
    ----------

    arguments:
        The given command line arguments

    """

    identifier = arguments.identifier
    measurement_time_s = arguments.time

    async with Network() as network:
        await network.connect_sensor_device(identifier)

        # Reduce sample rate to decrease CPU usage
        # See: https://github.com/MyTooliT/ICOc/issues/40
        adc_config = ADCConfiguration(
            reference_voltage=3.3,
            prescaler=2,
            acquisition_time=8,
            oversampling_rate=512,
        )
        await network.write_adc_configuration(**adc_config)

        sensor_range = await network.read_acceleration_sensor_range_in_g()
        conversion_to_g = partial(convert_raw_to_g, max_value=sensor_range)

        filepath = Path("Measurement.hdf5")

        with Storage(filepath.resolve()) as storage:
            sample_rate = (
                await network.read_adc_configuration()
            ).sample_rate()
            progress = tqdm(
                total=int(sample_rate * measurement_time_s),
                desc="Read sensor data",
                unit=" values",
                leave=False,
                disable=None,
            )

            async with network.open_data_stream(first=True) as stream:
                start_time = time()
                async for data in stream:
                    data.apply(conversion_to_g)
                    storage.add_streaming_data(data)
                    progress.update(3)  # 3 values per message

                    if time() - start_time >= measurement_time_s:
                        break
            storage.add_acceleration_meta(
                "Sensor_Range", f"± {sensor_range/2} g₀"
            )

            progress.close()


async def rename(arguments: Namespace) -> None:
    """Rename a sensor device

    Parameters
    ----------

    arguments:
        The given command line arguments

    """

    identifier = arguments.identifier
    name = arguments.name

    async with Network() as network:
        node = "STH 1"

        await network.connect_sensor_device(identifier)
        old_name = await network.get_name(node)
        mac_address = await network.get_mac_address(node)

        await network.set_name(name, node)
        name = await network.get_name(node)
        print(
            f"Renamed sensor device “{old_name}” with MAC "
            f"address “{mac_address}” to “{name}”"
        )


async def stu(arguments: Namespace) -> None:
    """Run specific commands regarding stationary transceiver unit

    Parameters
    ----------

    arguments:
        The given command line arguments

    """

    subcommand = arguments.stu_subcommand

    async with Network() as network:
        if subcommand == "ota":
            # The coroutine below activates the advertisement required for the
            # Over The Air (OTA) firmware update.
            #
            # - The `deactivate_bluetooth` command called when the execution
            #   leaves the `async with` block seems to not turn off the
            #   advertisement for the STU.
            # - Even a **hard STU reset does not turn off the advertisement**.
            # - One way to turn off the advertisement seems to be to initiate a
            #   connection with a sensor device.
            await network.activate_bluetooth()
        elif subcommand == "mac":
            print(await network.get_mac_address("STU 1"))
        elif subcommand == "reset":
            await network.reset_node("STU 1")
        else:
            raise ValueError(f"Unknown STU subcommand “{subcommand}”")


def main():
    """ICOtronic command line tool"""

    arguments = parse_arguments()
    basicConfig(
        level=arguments.log.upper(),
        style="{",
        format="{asctime} {levelname:7} {message}",
    )

    if arguments.subcommand == "config":
        config(arguments.subcommand)
    else:
        command_to_coroutine = {
            "config": config,
            "dataloss": dataloss,
            "list": list_sensor_devices,
            "measure": measure,
            "rename": rename,
            "stu": stu,
        }

        try:
            run(command_to_coroutine[arguments.subcommand](arguments))
        except NetworkError as error:
            print(error)
        except KeyboardInterrupt:
            pass


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    main()



# -- Imports ------------------------------------------------------------------

from unittest import main as unittest_main

from mytoolit.can.node import Node
from mytoolit.cmdline.commander import Commander
from mytoolit.config import settings
from mytoolit.measurement.sensor import guess_sensor
from mytoolit.report import Report
from mytoolit.test.production import TestSensorNode
from mytoolit.test.unit import ExtendedTestRunner

# -- Classes ------------------------------------------------------------------


class TestSMH(TestSensorNode):
    """This class contains tests for the milling head sensor (PCB)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()
        cls.report = Report(node="SMH")
        # Guessed sensor types
        cls.sensors = []
        for sensor in range(settings.smh.channels):
            cls.add_attribute(
                f"Sensor {sensor+1}", f"{{cls.sensors[{sensor}]}}", pdf=True
            )

    def _connect(self):
        """Create a connection to the SMH"""

        super()._connect_device(settings.smh.name)

    def _read_data(self):
        """Read data from connected SMH"""

        super()._read_data()

        cls = type(self)
        cls.name = settings.smh.name

    def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into SMH

        Please note the additional underscore in the method name that makes
        sure this test case is executed before all other test cases.

        The text `disconnected` in the method name make sure that the test
        framework does not initialize a connection.

        """

        self._test_firmware_flash(
            node="SMH",
            flash_location=settings.smh.firmware.location.flash,
            programmmer_serial_number=(
                settings.smh.programming_board.serial_number
            ),
            chip="BGM121A256V2",
        )

    def test_connection(self):
        """Check connection to SMH"""

        super()._test_connection_device()

    def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        async def test_eeprom():
            """Test the EERPOM of the SMH"""

            cls = type(self)
            receiver = Node("STH 1")

            # ========
            # = Name =
            # ========

            name = settings.smh.name

            await self.can.write_eeprom_name(name, receiver)
            read_name = await self.can.read_eeprom_name(receiver)

            self.assertEqual(
                name,
                read_name,
                (
                    f"Written name “{name}” does not match read name"
                    f" “{read_name}”"
                ),
            )

            cls.name = read_name

            # =========================
            # = Sleep & Advertisement =
            # =========================

            await self._test_eeprom_sleep_advertisement_times()

            # ================
            # = Product Data =
            # ================

            await self._test_eeprom_product_data(receiver, settings.smh)

            # ==============
            # = Statistics =
            # ==============

            await self._test_eeprom_statistics(
                receiver,
                settings.smh.production_date,
                settings.smh.batch_number,
            )

        self.loop.run_until_complete(test_eeprom())

    def test_sensors(self):
        """Test available sensor channels"""

        async def test_sensors():
            cls = type(self)

            for test_channel in range(1, settings.smh.channels + 1):
                await self.can.write_sensor_configuration(first=test_channel)
                config = await self.can.read_sensor_configuration()
                self.assertEqual(
                    config.first,
                    test_channel,
                    (
                        f"Read sensor channel number “{config.first}” does "
                        f"not match expected channel number “{test_channel}”"
                    ),
                )
                stream_data = await self.can.read_streaming_data_amount(
                    1000, first=True, second=False, third=False
                )
                values = [
                    timestamped.value for timestamped in stream_data.first
                ]
                sensor = guess_sensor(values)
                cls.sensors.append(sensor)

            non_working_sensors = [
                str(sensor_number)
                for sensor_number, sensor in enumerate(cls.sensors, start=1)
                if not sensor.works()
            ]

            if len(non_working_sensors) >= 1:
                if len(non_working_sensors) == 1:
                    error_text = f"channel {non_working_sensors.pop()} seems"
                elif len(non_working_sensors) >= 2:
                    channels = (
                        ", ".join(non_working_sensors[:-1])
                        + f" & {non_working_sensors[-1]}"
                    )
                    error_text = f"channels {channels} seem"
                plural = "" if len(non_working_sensors) <= 1 else "s"
                self.assertFalse(
                    non_working_sensors,
                    (
                        f"The sensor{plural} on measurement {error_text} "
                        "to not work correctly."
                    ),
                )

        self.loop.run_until_complete(test_sensors())

    def test_power_usage_disconnected(self) -> None:
        """Check power usage in disconnected state"""

        commander = Commander(
            serial_number=settings.smh.programming_board.serial_number,
            chip="BGM121A256V2",
        )

        commander.enable_debug_mode()
        power_usage_mw = commander.read_power_usage()

        expected_maxmimum_usage_mw = 10
        self.assertLess(
            power_usage_mw,
            expected_maxmimum_usage_mw,
            (
                f"Measured power usage of {power_usage_mw} mW is "
                "higher than expected maximum value "
                f"{expected_maxmimum_usage_mw} mW"
            ),
        )


# -- Main ---------------------------------------------------------------------


def main():
    unittest_main(
        testRunner=ExtendedTestRunner, module="mytoolit.test.production.smh"
    )


if __name__ == "__main__":
    main()



import unittest
import sys
import os

# Required to add peakcan
sDirName = os.path.dirname("")
sys.path.append(sDirName)
file_path = "../"
sDirName = os.path.dirname(file_path)
sys.path.append(sDirName)

import time

from mytoolit.old.network import Network
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr
from mytoolit.old.SthLimits import SthLimits
from mytoolit.old.MyToolItSth import TestConfig, fAdcRawDat
from mytoolit.old.MyToolItCommands import (
    ActiveState,
    AdcReference,
    byte_list_to_int,
    CalibMeassurementActionNr,
    CalibMeassurementTypeNr,
    DataSets,
    MyToolItBlock,
    MyToolItConfiguration,
    MyToolItStatData,
    MyToolItStreaming,
    MyToolItSystem,
    NodeState,
    NetworkState,
    payload2Hex,
)

sLogLocation = "../../Logs/STH/"

iSensorAxis = 1
bBatteryExternalDcDc = True
uAdc2Acc = 100
iRssiMin = -75
"""
This class supports a manual tests of the Sensory Tool Holder (STH)
"""


class TestSthManually(unittest.TestCase):
    def setUp(self):
        print("TestCase: ", self._testMethodName)
        input("Press Any Key to Continue")
        self.tSthLimits = SthLimits(
            iSensorAxis, bBatteryExternalDcDc, uAdc2Acc, iRssiMin, 20, 35
        )
        self.fileName = sLogLocation + self._testMethodName + ".txt"
        self.fileNameError = (
            sLogLocation + "Error_" + self._testMethodName + ".txt"
        )
        self.Can = Network(
            sender=MyToolItNetworkNr["SPU1"],
            receiver=MyToolItNetworkNr["STH1"],
            prescaler=self.tSthLimits.uSamplingRatePrescalerReset,
            acquisition=self.tSthLimits.uSamplingRateAcqTimeReset,
            oversampling=self.tSthLimits.uSamplingRateOverSamplesReset,
        )
        self.Can.logger.info("TestCase: " + str(self._testMethodName))
        self._resetStu()
        self.Can.logger.info("Connect to STH")
        self.Can.bBlueToothConnectPollingName(
            MyToolItNetworkNr["STU1"], TestConfig["DevName"]
        )
        self._resetSth()
        self.Can.logger.info("Connect to STH")
        self.Can.bBlueToothConnectPollingName(
            MyToolItNetworkNr["STU1"], TestConfig["DevName"]
        )
        self.bError = False
        self.Can.logger.info(
            "STU BlueTooth Address: "
            + hex(self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"]))
        )
        self.Can.logger.info(
            "STH BlueTooth Address: "
            + hex(self.Can.BlueToothAddress(MyToolItNetworkNr["STH1"]))
        )
        self._statusWords()
        temp = self._SthAdcTemp()
        self.assertGreaterEqual(self.tSthLimits.iTemperatureInternalMax, temp)
        self.assertLessEqual(self.tSthLimits.iTemperatureInternalMin, temp)
        print("Start")
        self.Can.logger.info("Start")

    def tearDown(self):
        if False != self.Can.bError:
            self.bError = True
        self.Can.__exit__()
        if self._test_has_failed():
            if os.path.isfile(self.fileNameError) and os.path.isfile(
                self.fileName
            ):
                os.remove(self.fileNameError)
            if os.path.isfile(self.fileName):
                os.rename(self.fileName, self.fileNameError)

    """
    Checks if test case has failed
    """

    def _test_has_failed(self):
        for _method, error in self._outcome.errors:
            if error:
                return True
        if True == self.bError:
            return True
        return False

    """
    Reset the STU
    """

    def _resetStu(self, retries=5, log=True):
        self.Can.bConnected = False
        return self.Can.reset_node("STU1", retries=retries, log=log)

    """
    Reset the STH
    """

    def _resetSth(self, retries=5, log=True):
        self.Can.bConnected = False
        return self.Can.reset_node("STH1", retries=retries, log=log)

    """
    Get the internal BGM113 Chip temeprature in °C
    """

    def _SthAdcTemp(self):
        ret = self.Can.calibMeasurement(
            MyToolItNetworkNr["STH1"],
            CalibMeassurementActionNr["Measure"],
            CalibMeassurementTypeNr["Temp"],
            1,
            AdcReference["1V25"],
            log=False,
        )
        result = float(byte_list_to_int(ret[4:]))
        result /= 1000
        self.Can.logger.info("Temperature(Chip): " + str(result) + "°C")
        self.Can.calibMeasurement(
            MyToolItNetworkNr["STH1"],
            CalibMeassurementActionNr["None"],
            CalibMeassurementTypeNr["Temp"],
            1,
            AdcReference["VDD"],
            log=False,
            bReset=True,
        )
        return result

    """
    Get all status words of STH and STU
    """

    def _statusWords(self):
        psw0 = self.Can.node_status(MyToolItNetworkNr["STH1"])
        self.Can.logger.info("STH Status Word: " + hex(psw0))
        psw0 = self.Can.node_status(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("STU Status Word: " + hex(psw0))

        status = self.Can.error_status(MyToolItNetworkNr["STH1"])
        if status.adc_overrun():
            self.bError = True
        self.Can.logger.info(f"STH Error Word: {status}")

        self.Can.logger.info(
            "STU Error Word: {}".format(
                self.Can.error_status(MyToolItNetworkNr["STU1"])
            )
        )

    """
    Turn off STH LED
    """

    def TurnOffLed(self):
        self.Can.logger.info("Turn Off LED")
        cmd = self.Can.CanCmd(
            MyToolItBlock["Configuration"], MyToolItConfiguration["Hmi"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [129, 1, 2, 0, 0, 0, 0, 0],
        )
        self.Can.tWriteFrameWaitAckRetries(message)

    """
    Turn on STH LED
    """

    def TurnOnLed(self):
        self.Can.logger.info("Turn On LED")
        cmd = self.Can.CanCmd(
            MyToolItBlock["Configuration"], MyToolItConfiguration["Hmi"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [129, 1, 1, 0, 0, 0, 0, 0],
        )
        self.Can.tWriteFrameWaitAckRetries(message)

    """
    Test Acknowledgement from STH. Write message and check identifier to be ack (No bError)
    """

    def testManually0001Ack(self):
        activeState = ActiveState()
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        msg = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["SPU1"], MyToolItNetworkNr["STH1"], [0]
        )
        self.Can.logger.info("Write Message")
        self.Can.WriteFrame(msg)
        self.Can.logger.info("Wait 200ms")
        time.sleep(0.2)
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 0, 0
        )
        msgAckExpected = self.Can.CanMessage20(
            cmd, MyToolItNetworkNr["STH1"], MyToolItNetworkNr["SPU1"], [0]
        )
        activeState.asbyte = self.Can.getReadMessage(-1).DATA[0]
        self.Can.logger.info(
            "Send ID: "
            + hex(msg.ID)
            + "; Expected ID: "
            + hex(msgAckExpected.ID)
            + "; Received ID: "
            + hex(self.Can.getReadMessage(-1).ID)
        )
        self.Can.logger.info(
            "Send Data: "
            + hex(0)
            + "; Received Data: "
            + hex(activeState.asbyte)
        )
        self.assertEqual(
            hex(msgAckExpected.ID), hex(self.Can.getReadMessage(-1).ID)
        )
        self.assertEqual(activeState.b.bSetState, 0)
        self.assertEqual(activeState.b.bReserved, 0)
        self.assertEqual(activeState.b.u2NodeState, NodeState["Application"])
        self.assertEqual(activeState.b.bReserved1, 0)
        self.assertEqual(
            activeState.b.u3NetworkState, NetworkState["Operating"]
        )

    """
    Test Standby Command - Run this manually! Only (power on) reset gets out of that (or replay firmware)
    """

    def testManually0010Standby(self):
        failTry = ActiveState()
        receivedData = ActiveState()
        receivedDataFailTry = ActiveState()
        sendData = ActiveState()
        sendData.asbyte = 0
        sendData.b.bSetState = 1
        sendData.b.u2NodeState = 2
        sendData.b.u3NetworkState = 2
        failTry.asbyte = 0
        failTry.b.bSetState = 1
        failTry.b.u2NodeState = 2
        failTry.b.u3NetworkState = 6
        cmd = self.Can.CanCmd(
            MyToolItBlock["System"], MyToolItSystem["Get/Set State"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [sendData.asbyte],
        )
        self.Can.logger.info("Send Shut Down Command")
        receivedData.asbyte = self.Can.tWriteFrameWaitAckRetries(message)[
            "Payload"
        ][0]
        self.Can.logger.info("Send try should fail")
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [failTry.asbyte],
        )
        self.Can.WriteFrame(message)
        self.Can.logger.info("Wait 200ms")
        time.sleep(0.2)
        receivedDataFailTry.asbyte = self.Can.getReadMessage(-1).DATA[0]
        self.Can.logger.info(
            "Fail Try Payload Byte for Active State Command(send): "
            + str(sendData.asbyte)
        )
        self.Can.logger.info(
            "Fail Try Payload Byte for Active State Command(Received): "
            + str(receivedData.asbyte)
        )
        self.Can.logger.info(
            "Fail Try Payload Byte for Active State Command(send with fail): "
            + str(failTry.asbyte)
        )
        self.Can.logger.info(
            "Fail Try Payload Byte for Active State Command(last received): "
            + str(receivedDataFailTry.asbyte)
        )
        self.assertEqual(receivedData.asbyte, sendData.asbyte)
        self.assertEqual(receivedData.asbyte, receivedDataFailTry.asbyte)
        print(
            "Power off device for 1 minute(power consumpiton of the target is"
            " actually REALLY low)"
        )
        input("Press any key to continue")

    """
    Power Consumption - Energy Save Modes
    """

    def testManually0012PowerConsumptionStandby(self):
        self.Can.Standby(MyToolItNetworkNr["STH1"])
        print("Start Simplicty Energy Profiler and connect to target (STH)")
        print("Measure Power Consumption for standby.")
        input("Press any key to continue")
        print(
            "Power off device for 1 minute(power consumpiton of the target is"
            " actually REALLY low)"
        )
        input("Press any key to continue")

    def vComapre(self, iCounterCompare, aiAccCounter, iIndex):
        if iCounterCompare != aiAccCounter[iIndex]:
            iErrorIndex = iIndex
            self.Can.logger.error("Error starts at index: " + str(iIndex))
            if 255 <= iIndex:
                iIndex -= 255
            else:
                iIndex = 0
            iIndexEnd = 255 * 4 + 1
            while iIndex + iIndexEnd > len(aiAccCounter):
                iIndexEnd -= 1
            for i in range(iIndex, iIndex + iIndexEnd):
                self.Can.logger.error(str(aiAccCounter[i]))
            self.assertEqual(iCounterCompare, aiAccCounter[iErrorIndex])

    """
    Power Consumption - Energy Save Modes
    """

    def testManually0300MeasuringInterference(self):
        self.Can.ConfigAdc(
            MyToolItNetworkNr["STH1"],
            self.tSthLimits.uSamplingRatePrescalerReset,
            self.tSthLimits.uSamplingRateAcqTimeReset,
            self.tSthLimits.uSamplingRateOverSamplesReset,
            AdcReference["VDD"],
        )
        print(
            "Please take a smartphone, start scanning Bluetooth devices and"
            " hold it over STU and STH alternately for 10s"
        )
        [indexStart, indexEnd] = self.Can.streamingValueCollect(
            MyToolItNetworkNr["STH1"],
            MyToolItStreaming["Data"],
            DataSets[3],
            1,
            0,
            0,
            30000,
        )
        [
            arrayAccX,
            arrayAccY,
            arrayAccZ,
        ] = self.Can.streamingValueArrayMessageCounters(
            MyToolItNetworkNr["STH1"],
            MyToolItStreaming["Data"],
            DataSets[3],
            1,
            0,
            0,
            indexStart,
            indexEnd,
        )
        self.Can.ValueLog(
            arrayAccX, arrayAccY, arrayAccZ, fAdcRawDat, "AccMsgCounter", ""
        )
        self.assertEqual(0, len(arrayAccY))
        self.assertEqual(0, len(arrayAccZ))
        count = arrayAccX[0]
        self.assertGreaterEqual(count, 0)
        self.assertLessEqual(count, 255)
        bNok = False
        i = 0
        while i < int(len(arrayAccX) / 3):
            if count != arrayAccX[i]:
                bNok = True
            i += 1
            if count != arrayAccX[i]:
                bNok = True
            i += 1
            if count != arrayAccX[i]:
                bNok = True
            i += 1
            count += 1
            count %= 256
        self.assertEqual(bNok, True)

        self.assertGreaterEqual(count, 0)
        self.assertLessEqual(count, 255)
        i = int(2 * len(arrayAccX) / 3)
        while (
            arrayAccX[i] != arrayAccX[i + 1]
            or arrayAccX[i] != arrayAccX[i + 2]
        ):
            i += 1
        self.assertLess(i, int(2 * len(arrayAccX) / 3) + 3)
        count = arrayAccX[i]
        while i < len(arrayAccX) - 3:
            for _j in range(0, 3):
                self.vComapre(count, arrayAccX, i)
                i += 1
            count += 1
            count %= 256

    """
    Under Voltage Counter
    """

    def testManually0700UnderVoltageCounter(self):
        UnderVoltage1 = self.Can.statisticalData(
            MyToolItNetworkNr["STH1"], MyToolItStatData["Uvc"], printLog=True
        )
        UnderVoltagePowerOnFirst1 = byte_list_to_int(UnderVoltage1[:4])
        self.Can.logger.info(
            "Under Voltage Counter since first Power On: "
            + payload2Hex(UnderVoltage1)
        )
        self.Can.logger.info(
            "Under Voltage Counter since first Power On: "
            + str(UnderVoltagePowerOnFirst1)
        )
        input(
            "Power Off Device and wait 1s, power on again and then press Any"
            " Key to Continue"
        )
        self.Can.bBlueToothConnectPollingName(
            MyToolItNetworkNr["STU1"], TestConfig["DevName"]
        )
        UnderVoltage2 = self.Can.statisticalData(
            MyToolItNetworkNr["STH1"], MyToolItStatData["Uvc"], printLog=True
        )
        UnderVoltagePowerOnFirst2 = byte_list_to_int(UnderVoltage2[:4])
        self.Can.logger.info(
            "Under Voltage Counter since first Power On: "
            + payload2Hex(UnderVoltage2)
        )
        self.Can.logger.info(
            "Under Voltage Counter since first Power On: "
            + str(UnderVoltagePowerOnFirst2)
        )
        self.assertEqual(
            0xFFFFFFFF & (UnderVoltagePowerOnFirst1 + 1),
            UnderVoltagePowerOnFirst2,
        )


if __name__ == "__main__":
    print(sys.version)
    sLogLocation = sys.argv[1]
    sLogFile = sys.argv[2]
    if "/" != sLogLocation[-1]:
        sLogLocation += "/"
    sLogFileLocation = sLogLocation + sLogFile
    sDirName = os.path.dirname(sLogFileLocation)
    sys.path.append(sDirName)

    print("Log Files will be saved at: " + str(sLogFileLocation))
    if not os.path.exists(sDirName):
        os.makedirs(sDirName)
    with open(sLogFileLocation, "w") as f:
        print(f)
        runner = unittest.TextTestRunner(f)
        unittest.main(argv=["first-arg-is-ignored"], testRunner=runner)



# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from typing import Optional, Union

from mytoolit.old.MyToolItCommands import (
    MyToolItBlock,
    blocknumber_to_commands,
)

# -- Class --------------------------------------------------------------------


class Command:
    """This class represents a command including error and acknowledge bits"""

    def __init__(
        self,
        *command: int,
        block: Union[None, str, int] = None,
        block_command: Union[None, str, int] = None,
        request: Optional[bool] = None,
        error: Optional[bool] = None,
    ) -> None:
        """Create a new command from the given arguments

        Usually you will either specify the command directly, or provide
        block, block_command and values for the error and acknowledge/request
        bits. If you decide to specify both the command value and one of the
        keyword arguments, then the keyword arguments will be used to
        overwrite specific parts of the command. For more information, please
        take a look at the examples.


        Parameters
        ---------

        command:
            A 16 bit number that specifies the command including acknowledge
            and error bits (16 bit)

        block:
            A 6 bit number or string that specifies the block of the command

        block_command:
            A 8 bit number or string that specifies the block command

        request:
            A boolean value that specifies if the command is for a request or
            an acknowledgement

        error:
            A boolean value that defines if there was an error or not

        Examples
        --------

        >>> Command(block=0, block_command=0).value
        0

                        block   command A E
        >>> command = 0b001000_00000100_0_0
        >>> bin(Command(command).value)
        '0b10000000010000'

        >>> command = Command(block='System')
        >>> command.block_name()
        'System'

        >>> Command(block='Does Not Exist')
        Traceback (most recent call last):
            ...
        ValueError: Unknown block: Does Not Exist

        >>> command = Command(block='Streaming', block_command='Data')
        >>> command.block_command_name()
        'Data'

        >>> Command(block='Streaming', block_command='Does Not Exist')
        Traceback (most recent call last):
            ...
        ValueError: Unknown block command: Does Not Exist

        """

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            command_ones = 0xFFFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ command_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set command bits to given value
            self.value |= number << start

        # ===========
        # = Command =
        # ===========

        self.value = command[0] if command else 0

        # =========
        # = Block =
        # =========

        if isinstance(block, str):
            try:
                block = MyToolItBlock[block]
                assert isinstance(block, int)
            except KeyError:
                raise ValueError(f"Unknown block: {block}")

        if block is not None:
            set_part(start=10, width=6, number=block)

        # =================
        # = Block Command =
        # =================

        if isinstance(block_command, str):
            try:
                block_command_names = blocknumber_to_commands[self.block()]
            except KeyError:
                raise ValueError(f"Unknown block number: {block}")
            else:
                try:
                    block_command = block_command_names[block_command]
                except KeyError:
                    raise ValueError(f"Unknown block command: {block_command}")

        if block_command is not None:
            set_part(start=2, width=8, number=block_command)

        # ===================
        # = Request & Error =
        # ===================

        if request is not None:
            set_part(start=1, width=1, number=int(bool(request)))

        if error is not None:
            set_part(start=0, width=1, number=int(bool(error)))

    def __repr__(self) -> str:
        """Get a textual representation of the command

        Returns
        -------

        A string that describes the various attributes of the command

        Examples
        --------

                      block   command A E
        >>> Command(0b001000_00000100_0_0)
        Block: StatisticalData, Command: ProductionDate, Acknowledge

                      block   command A E
        >>> Command(block=0, block_command=0x0c, request=True, error=False)
        Block: System, Command: Routing, Request

        """
        error = self.value & 1

        attributes = filter(
            None,
            [
                f"Block: {self.block_name()}",
                f"Command: {self.block_command_name()}",
                "Acknowledge" if self.is_acknowledgment() else "Request",
                "Error" if error else None,
            ],
        )

        return ", ".join(attributes)

    def block(self) -> int:
        """Get the block

        Returns
        -------

        The block number of the command

        Example
        -------

                      block   command A E
        >>> Command(0b000011_00000000_0_0).block()
        3

        """

        return (self.value >> 10) & 0b111111

    def block_name(self) -> str:
        """Get a short description of the block

        Returns
        -------

        A short textual representation of the block number

        Examples
        --------
                      block   command A E
        >>> Command(0b101000_00000010_0_0).block_name()
        'Configuration'

                      block   command A E
        >>> Command(0b111101_00000010_0_0).block_name()
        'EEPROM'

        """

        return MyToolItBlock.inverse.get(self.block(), "Unknown")

    def block_command(self) -> int:
        """Get the block command number

        Returns
        -------

        The block command number of the command

        Example
        -------

                      block   command A E
        >>> Command(0b001000_00000100_0_0).block_command()
        4

        """

        return (self.value >> 2) & 0xFF

    def block_command_name(self) -> str:
        """Get the name of the block command

        Returns
        -------

        A short textual representation of the block command

        Examples
        --------

                      block   command A E
        >>> Command(0b101000_00000000_0_0).block_command_name()
        'Get/Set ADC Configuration'

                      block   command A E
        >>> Command(0b000000_00001011_0_0).block_command_name()
        'Bluetooth'

        """

        try:
            return blocknumber_to_commands[self.block()].inverse[
                self.block_command()
            ]
        except KeyError:
            return "Unknown"

    def is_acknowledgment(self) -> bool:
        """Checks if this command represents an acknowledgment

        Returns
        -------

        True if the command is for an acknowledgement, or false otherwise

        Examples
        --------

                      block   command A E
        >>> Command(0b101000_00000000_0_0).is_acknowledgment()
        True

        >>> Command(request=True).is_acknowledgment()
        False

        """

        return bool((self.value >> 1) & 1 == 0)

    def is_error(self) -> bool:
        """Checks if the command represents an error

        Returns
        -------

        True if the command represents an error, or false otherwise

        Examples
        --------
                      block   command A E
        >>> Command(0b101011_00000001_1_0).is_error()
        False

                      block   command A E
        >>> Command(0b101010_00000000_0_1).is_error()
        True

        """

        return bool(self.value & 1)

    def set_acknowledgment(self, value: bool = True) -> Command:
        """Set the acknowledgment bit to the given value

        Parameters
        ---------

        value:
            A boolean that specifies if the command represents an
            acknowledgment or not

        Examples
        --------

        >>> Command().set_acknowledgment().is_acknowledgment()
        True

        >>> Command().set_acknowledgment(True).is_acknowledgment()
        True

        >>> Command().set_acknowledgment(False).is_acknowledgment()
        False

        Returns
        -------

        The modified command object

        """

        request = not value
        request_bit = 1 << 1

        if request:
            self.value |= request_bit
        else:
            command_ones = 0xFF
            self.value &= request_bit ^ command_ones

        return self

    def set_error(self, error: bool = True) -> Command:
        """Set the error bit to the given value

        Parameters
        ----------

        error:
            A boolean that specifies if the command represents an error or not

        Examples
        --------

        >>> Command().set_error().is_error()
        True

        >>> Command().set_error(True).is_error()
        True

        >>> Command().set_error(False).is_error()
        False

        Returns
        -------

        The modified command object

        """

        if error:
            self.value |= 1
        else:
            self.value &= 0b11111110

        return self


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from functools import partial
from importlib.resources import as_file, files
from typing import List

from reportlab.lib.units import cm
from reportlab.platypus import (
    Flowable,
    ListFlowable,
    Paragraph,
    SimpleDocTemplate,
    Spacer,
    Table,
)
from reportlab.rl_config import defaultPageSize

from mytoolit.report.pdf import PDFImage
from mytoolit.report.style import get_style_sheet
from mytoolit.report.forms import CheckBoxList, KeepTogether, Signature

# -- Functions ----------------------------------------------------------------


# noinspection PyUnusedLocal
def _first_page(canvas, document, node):
    """Define the style of the first page of the report"""

    canvas.saveState()

    page_width = defaultPageSize[0]
    page_height = defaultPageSize[1]
    logo_width = 370
    logo_height = 75
    logo_offset = 50
    title_offset = logo_offset + logo_height + 20

    with as_file(files("mytoolit.report") / "MyTooliT.pdf") as logo_filepath:
        PDFImage(logo_filepath, logo_width, logo_height).drawOn(
            canvas,
            (page_width - logo_width) / 2,
            page_height - logo_offset - logo_height,
        )

    style = get_style_sheet()

    center_width = page_width / 2

    heading1 = style["Heading1"]
    canvas.setFont(heading1.fontName, heading1.fontSize)
    canvas.drawCentredString(
        center_width, page_height - title_offset, f"{node} Test Report"
    )

    canvas.restoreState()


# -- Class --------------------------------------------------------------------


class Report:
    """Generate test reports using ReportLab"""

    story: List[Flowable]  # Improve happiness of PyCharm type checker

    def __init__(self, node="Device"):
        """Initialize the report

        Parameters
        ----------

        node:
            A text that specifies the node (e.g. STH) for which a report
            should be generated

        """

        self.node = node
        self.document = SimpleDocTemplate(
            f"{node} Test.pdf",
            author="MyTooliT",
            title="Test Report",
            subject="{} Test".format(
                "Sensory Tool Holder"
                if node == "STH"
                else (
                    "Sensory Milling Head"
                    if node == "SMH"
                    else "Stationary Transceiver Unit"
                )
            ),
        )
        self.story = [Spacer(1, 3 * cm)]
        self.styles = get_style_sheet()

        self.general = []
        self.attributes = []
        self.tests = []
        self.checks = []

    def __add_header(self, text):
        """Add a header at the current position in the document

        Parameters
        ----------

        text:
            The text of the heading

        """

        self.story.append(Paragraph(text, style=self.styles["Heading2"]))

    def __add_table(self, data, column_widths=None):
        """Add a table at the current position in the document

        Parameters
        ----------

        data:
            The data that should be stored in the table

        column_widths:
            The width of each column of the table

        """

        table = Table(data, colWidths=column_widths)
        table.hAlign = "LEFT"
        self.story.append(table)

    def add_attribute(self, name, value, node_attribute=True):
        """Add information about an attribute to the report

        Parameters
        ----------

        name:
            The name of the attribute

        value:
            The value of the attribute

        sth_attribute

            Specifies if the specified name and value stores node specific data
            or general data

        """

        table = self.attributes if node_attribute else self.general
        table.append([name, value])

    def add_test_result(self, description, result):
        """Add information about a single test result to the report

        Parameters
        ----------

        description:
            A textual description of the test

        result:
            The unit test result of the test

        """

        test = result.last_test

        result_text = (
            "<font color='red'>Error</font>"
            if test.error()
            else (
                "<font color='orange'>Failure</font>"
                if test.failure()
                else "<font color='green'>Ok</font>"
            )
        )

        result_text = f"{description}: <b>{result_text}</b>"
        if test.message:
            test.message = f"{test.message}".replace("\n", "<br/>")
            result_text += f"<br/><br/><b>{test.message}</b><br/><br/>"
        paragraph_result = Paragraph(result_text, style=self.styles["Normal"])
        self.tests.append(paragraph_result)

    def add_checkbox_list(
        self, title: str, boxes: List[str], text_fields: int = 0
    ) -> None:
        """Add a checkbox list to the report

        Parameters
        ----------

        title:
            The title that should be printed before the checkbox list

        boxes:
            A text for each box that should be added to the checkbox list

        text_fields:
            The number of text fields for additional comments that should be
            added at the end of the checklist.

        """

        checkbox_list = CheckBoxList(title, text_fields)

        for box in boxes:
            checkbox_list.add_checkbox_item(box)

        self.checks.append(checkbox_list)

    def build(self):
        """Store the PDF report"""

        if len(self.general) > 0:
            self.__add_header("General")
            self.__add_table(self.general)

        if len(self.attributes) > 0:
            self.__add_header("Attributes")
            self.__add_table(self.attributes)

        if len(self.tests) > 0:
            self.__add_header("Test Results")
            tests = ListFlowable(self.tests, bulletType="bullet")
            self.story.append(tests)

        if len(self.checks) > 0:
            title = Paragraph("Manual Checks", style=self.styles["Heading2"])
            check_lists = [
                checkbox_list.to_flowable() for checkbox_list in self.checks
            ]

            # We try to keep title and the first checklist on the same page
            # Unfortunately, this currently means that all checklists will be
            # added at the same page, even if there would be enough space for
            # some of the checklists on the previous page. We tried to address
            # this by setting the `keepWithNext` attribute in the `Paragraph`
            # and `ParagraphStyle`. Unfortunately this does not change the flow
            # at all, which means the header “Manual Checks” might be added
            # at the end of a page without any following additional content on
            # the same page :(.

            self.story.append(KeepTogether([title, check_lists[0]]))
            self.story.extend(check_lists[1:])

        self.story.append(Signature().to_flowable())

        first_page = partial(_first_page, node=self.node)
        self.document.build(self.story, onFirstPage=first_page)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from platform import system
    from subprocess import run

    # Create a example test report
    node = "STH"
    pdf_path = f"{node} Test.pdf"
    report = Report(node)

    report.add_checkbox_list(
        title="Checkbox Title", boxes=["First box", "Second box"]
    )

    report.build()

    # Open the file to check the resulting PDF manually
    if system() == "Windows":
        # The function `startfile` is only available on Windows
        from os import startfile  # type: ignore

        startfile(pdf_path)
    else:
        run(
            ["{}open".format("" if system() == "Darwin" else "xdg-"), pdf_path]
        )



from mytoolit.old.MyToolItCommands import (
    AdcAcquisitionTime,
    AdcOverSamplingRate,
    AdcReference,
    calcSamplingRate,
    AdcMax,
)


class SthLimits:
    def __init__(
        self,
        iAccSensorAxis,
        uAccelerationToAccGravitity,
        iTemperatureInternalMin,
        iTemperatureInternalMax,
        battery=True,
    ):
        self.vBatteryParameters(battery)
        self.vAccParameters(iAccSensorAxis, uAccelerationToAccGravitity)
        self.vSamplingRate()
        self.vStreaming()
        self.vVoltageRaw()
        self.vTemperature(iTemperatureInternalMin, iTemperatureInternalMax)
        self.vVirtualFullScale()
        self.vAdcBuffer(3, 24)

    def vBatteryParameters(self, bConnectedBattery):
        if False != bConnectedBattery:
            self.uBatteryMiddle = 3.9
            self.uBatteryQ1 = 8000
            self.uBatteryQ25 = 9000
            self.uBatteryMedL = 10000
            self.uBatteryMedH = 12000
            self.uBatteryQ75 = 13000
            self.uBatteryQ99 = 14000
            self.uBatteryVar = 40000
            self.uBatterySkewness = 20
            self.uBatterySNR = 60
            self.uBatteryMiddleRaw = 11000
            self.uBatteryToleranceRaw = 300
        else:
            self.uBatteryMiddle = 3.3
            self.uBatteryQ1 = 600
            self.uBatteryQ25 = 630
            self.uBatteryMedL = 633
            self.uBatteryMedH = 670
            self.uBatteryQ75 = 680
            self.uBatteryQ99 = 710
            self.uBatteryVar = 40000
            self.uBatterySkewness = 20
            self.uBatterySNR = 60
            self.uBatteryMiddleRaw = 60
            self.VoltRawMiddleBat = 700
            self.uBatteryToleranceRaw = 200
        self.uBatteryTolerance = 0.3
        self.iBatteryK = (57 * 3.3) / (10 * 2**16)

    def auBatteryStatisticsHist(self):
        return [
            self.uBatteryQ1,
            self.uBatteryQ25,
            self.uBatteryMedL,
            self.uBatteryMedH,
            self.uBatteryQ75,
            self.uBatteryQ99,
        ]

    def auBatteryStatisticsMoment(self):
        return [self.uBatteryVar, self.uBatterySkewness, self.uBatterySNR]

    def vAccParameters(self, iAccSensorAxis, uAccelerationToAccGravitity):
        self.uAccelerationToAccGravitity = uAccelerationToAccGravitity
        self.iAdcAccXMiddle = 0
        self.iAdcAccYMiddle = 0
        self.iAdcAccZMiddle = 0
        self.iAdcAccXTolerance = 3.5
        self.iAdcAccXRawMiddle = 2**15
        self.iAdcAccXRawTolerance = 2048
        self.uAccXQ1 = 2**15 * 0.95
        self.uAccXQ25 = 2**15 * 0.955
        self.uAccXMedL = 2**15 * 0.97
        self.uAccXMedH = 2**15 * 1.03
        self.uAccXQ75 = 2**15 * 1.045
        self.uAccXQ99 = 2**15 * 1.05
        self.uAccXVar = 2000
        self.uAccXSkewness = 0.9
        self.uAccXSNR = -60
        if 1 == iAccSensorAxis:
            self.iAdcAccYTolerance = 200
            self.iAdcAccZTolerance = 200
            self.iAdcAccYRawMiddle = 0
            self.iAdcAccZRawMiddle = 0
            self.iAdcAccYRawTolerance = 54000
            self.iAdcAccZRawTolerance = 54000
            self.uAccYQ1 = 0
            self.uAccYQ25 = 1
            self.uAccYMedL = 2
            self.uAccYMedH = 2**16 - 3
            self.uAccYQ75 = 2**16 - 2
            self.uAccYQ99 = 2**16 - 1
            self.uAccYVar = 40000
            self.uAccYSkewness = 20
            self.uAccYSNR = 60
            self.uAccZQ1 = 0
            self.uAccZQ25 = 1
            self.uAccZMedL = 2
            self.uAccZMedH = 2**16 - 3
            self.uAccZQ75 = 2**16 - 2
            self.uAccZQ99 = 2**16 - 1
            self.uAccZVar = 40000
            self.uAccZSkewness = 20
            self.uAccZSNR = 60
        else:
            self.iAdcAccYTolerance = 3
            self.iAdcAccZTolerance = 3
            self.iAdcAccYRawMiddle = 2**15
            self.iAdcAccZRawMiddle = 2**15
            self.iAdcAccYRawTolerance = 2048
            self.iAdcAccZRawTolerance = 2048
            self.uAccYQ1 = 2**15 * 0.95
            self.uAccYQ25 = 2**15 * 0.955
            self.uAccYMedL = 2**15 * 0.97
            self.uAccYMedH = 2**15 * 1.03
            self.uAccYQ75 = 2**15 * 1.045
            self.uAccYQ99 = 2**15 * 1.05
            self.uAccYVar = 2000
            self.uAccYSkewness = 0.9
            self.uAccYSNR = 60
            self.uAccZQ1 = 2**15 * 0.95
            self.uAccZQ25 = 2**15 * 0.955
            self.uAccZMedL = 2**15 * 0.97
            self.uAccZMedH = 2**15 * 1.03
            self.uAccZQ75 = 2**15 * 1.045
            self.uAccZQ99 = 2**15 * 1.05
            self.uAccZVar = 2000
            self.uAccZSkewness = 0.9
            self.uAccZSNR = 60
        self.iSelfTestOutputChangemVMin = 70
        self.iSelfTestOutputChangemVTyp = 170
        self.iAccX_K = self.uAccelerationToAccGravitity / (2**16 - 1)
        self.iAccX_D = -(self.uAccelerationToAccGravitity / 2)

    def auAccXStatisticsHist(self):
        return [
            self.uAccXQ1,
            self.uAccXQ25,
            self.uAccXMedL,
            self.uAccXMedH,
            self.uAccXQ75,
            self.uAccXQ99,
        ]

    def auAccXStatisticsMoment(self):
        return [self.uAccXVar, self.uAccXSkewness, self.uAccXSNR]

    def auAccYStatisticsHist(self):
        return [
            self.uAccYQ1,
            self.uAccYQ25,
            self.uAccYMedL,
            self.uAccYMedH,
            self.uAccYQ75,
            self.uAccYQ99,
        ]

    def auAccYStatisticsMoment(self):
        return [self.uAccYVar, self.uAccYSkewness, self.uAccYSNR]

    def auAccZStatisticsHist(self):
        return [
            self.uAccZQ1,
            self.uAccZQ25,
            self.uAccZMedL,
            self.uAccZMedH,
            self.uAccZQ75,
            self.uAccZQ99,
        ]

    def auAccZStatisticsMoment(self):
        return [self.uAccZVar, self.uAccZSkewness, self.uAccZSNR]

    def fAcceleration(self, x):
        return (x / AdcMax - 1 / 2) * self.uAccelerationToAccGravitity

    def vSamplingRate(self):
        self.uSamplingRatePrescalerReset = 2
        self.uSamplingRateAcqTimeReset = AdcAcquisitionTime[8]
        self.uSamplingRateOverSamplesReset = AdcOverSamplingRate[64]
        self.uSamplingRateMin = 150  # To check maximum frequencies to be max
        self.uSamplingRateMax = 14000  # To check maximum frequencies to be max
        self.uSamplingRateSinglePrescalerMax = 2
        self.uSamplingRateSingleAcqTimeMax = AdcAcquisitionTime[8]
        self.uSamplingRateSingleOverSamplesMax = AdcOverSamplingRate[64]
        self.uSamplingRateDoublePrescalerMax = 2
        self.uSamplingRateDoubleAcqTimeMax = AdcAcquisitionTime[16]
        self.uSamplingRateDoubleOverSamplesMax = AdcOverSamplingRate[64]
        self.uSamplingRateTripplePrescalerMax = 2
        self.uSamplingRateTrippleAcqTimeMax = AdcAcquisitionTime[8]
        self.uSamplingRateTrippleOverSamplesMax = AdcOverSamplingRate[64]
        self.uSamplingToleranceLow = 0.97
        self.uSamplingToleranceHigh = 1.03

    def uSamplingRateSingleMax(self):
        return calcSamplingRate(
            self.uSamplingRateSinglePrescalerMax,
            self.uSamplingRateSingleAcqTimeMax,
            self.uSamplingRateSingleOverSamplesMax,
        )

    def uSamplingRateDoubleMax(self):
        return calcSamplingRate(
            self.uSamplingRateDoublePrescalerMax,
            self.uSamplingRateDoubleAcqTimeMax,
            self.uSamplingRateDoubleOverSamplesMax,
        )

    def uSamplingRateTrippleMax(self):
        return calcSamplingRate(
            self.uSamplingRateTripplePrescalerMax,
            self.uSamplingRateTrippleAcqTimeMax,
            self.uSamplingRateTrippleOverSamplesMax,
        )

    def vStreaming(self):
        self.uStartupTimeMs = 250
        self.uStandardTestTimeMs = 10000

    def vVoltageRaw(self):
        self.uVoltRawOpa2Middle = 0
        self.uVoltRawOpa2Tolerance = 8
        self.uVoltRawOpa3Middle = 0
        self.uVoltRawOpa3Tolerance = 8
        self.uVoltRawVssTolerance = 8

    def vTemperature(self, iTemperatureInternalMin, iTemperatureInternalMax):
        self.uTemperatureInternal3V3Middle = 13000
        self.uTemperatureInternal3V3Tolerance = 2000
        self.iTemperatureInternalMin = iTemperatureInternalMin
        self.iTemperatureInternalMax = iTemperatureInternalMax

    def vVirtualFullScale(self):
        self.Vfs = 32

    def vAdcBuffer(self, sizeX, sizeY):
        self.uAdcSizeX = sizeX
        self.uAdcSizeY = sizeY

    def uAdcBufferSizeBytes(self):
        return self.uAdcSizeX * self.uAdcSizeY



# -- Imports ------------------------------------------------------------------

from itertools import chain, repeat
from typing import List
from unittest import main as unittest_main, skipIf

from semantic_version import Version

from mytoolit.can import Node
from mytoolit.config import settings
from mytoolit.measurement import convert_raw_to_g, g0, ratio_noise_max, volt
from mytoolit.report import Report
from mytoolit.test.production import TestSensorNode
from mytoolit.test.unit import ExtendedTestRunner
from mytoolit.utility import convert_mac_base64

# -- Classes ------------------------------------------------------------------


class TestSTH(TestSensorNode):
    """This class contains tests for the Sensory Tool Holder (STH)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()

        cls.add_attribute("Holder Type", "{cls.holder_type}", pdf=True)
        cls.add_attribute(
            "Acceleration Sensor", "{cls.acceleration_sensor}", pdf=True
        )

        for axis in "xyz":
            cls.add_attribute(
                f"Acceleration Slope {axis.upper()}",
                "{cls.acceleration_slope_" + axis + ":.5f}",
                pdf=False,
            )
            cls.add_attribute(
                f"Acceleration Offset {axis.upper()}",
                "{cls.acceleration_offset_" + axis + ":.3f}",
                pdf=False,
            )

        cls.report = Report(node="STH")

        # Add data that only applies to the STH
        cls.holder_type = settings.sth.holder_type
        cls.status = settings.sth.status

        sensor_name = settings.sth.acceleration_sensor.sensor
        maximum_acceleration = (
            settings.acceleration_sensor().acceleration.maximum
        )
        cls.acceleration_sensor = (
            f"±{maximum_acceleration//2} g Sensor ({sensor_name})"
        )

        # Manual checks
        cls.report.add_checkbox_list(
            title="Metal Blank",
            boxes=[
                "Okay",
                "Cylindrical thread defect",
                "Dent",
                "Oil spillage",
                "Shavings",
                "Milling errors",
            ],
            text_fields=1,
        )

        cls.report.add_checkbox_list(
            title="PCB",
            boxes=["Optical inspection: no defects"],
            text_fields=1,
        )

        cls.report.add_checkbox_list(
            title="Before Resin Cast",
            boxes=[
                "Battery test successful",
                "Charge in charging station was successful",
            ],
            text_fields=2,
        )

        cls.report.add_checkbox_list(
            title="Final Checks",
            boxes=[
                "Resin cast contains no bubbles",
                "Resin cast hardened completely",
                "No resin residue outside of pocket",
                "Pocket is completely filled with resin",
                "No oil spillage in vacuum chamber",
                "Charge in charging station was successful",
            ],
            text_fields=2,
        )

    def setUp(self):
        """Set up hardware before a single test case"""

        async def write_adc_configuration():
            """Change reference voltage depending on acceleration sensor"""

            reference_voltage = (
                settings.acceleration_sensor().reference_voltage
            )

            await self.can.write_adc_configuration(
                prescaler=2,
                acquisition_time=8,
                oversampling_rate=64,
                reference_voltage=reference_voltage,
            )

        super().setUp()

        # Sensor node is connected after set up function unless the test
        # does not initiate a Bluetooth connection, which is the case if
        # the test name (method name) contains the text “disconnected”.
        if self._testMethodName.find("disconnected") < 0:
            self.loop.run_until_complete(write_adc_configuration())

    def _connect(self):
        """Create a connection to the STH"""

        super()._connect_device(settings.sth_name())

    def _read_data(self):
        """Read data from connected STH"""

        super()._read_data()

        cls = type(self)
        cls.name = settings.sth_name()

    @skipIf(
        settings.sth.status == "Epoxied",
        f"Flash test skipped because of status “{settings.sth.status}”",
    )
    def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into STH

        Please note the additional underscore in the method name that makes
        sure this test case is executed before all other test cases.

        The text `disconnected` in the method name makes sure that the test
        framework does not initialize a connection.

        """

        try:
            hardware_version = Version(settings.sth.hardware_version)
        except (TypeError, ValueError) as error:
            raise ValueError(
                "Incorrect STH hardware version: "
                f"“{settings.sth.hardware_version}”"
            ) from error

        if not (1 <= hardware_version.major <= 2):
            raise ValueError(
                f"STH hardware version “{hardware_version}” "
                "is currently not supported"
            )

        chip = (
            "BGM113A256V2" if hardware_version.major <= 1 else "BGM123A256V2"
        )

        self._test_firmware_flash(
            node="STH",
            flash_location=settings.sth.firmware.location.flash,
            programmmer_serial_number=(
                settings.sth.programming_board.serial_number
            ),
            chip=chip,
        )

    def test_connection(self):
        """Check connection to STH"""

        super()._test_connection_device()

    def test_battery_voltage(self):
        """Test voltage of STH power source"""

        async def test_supply_voltage():
            """Check the supply voltage of the STH"""

            supply_voltage = await self.can.read_supply_voltage()
            expected_voltage = volt(settings.sth.battery_voltage.average)
            tolerance_voltage = volt(settings.sth.battery_voltage.tolerance)
            expected_minimum_voltage = expected_voltage - tolerance_voltage
            expected_maximum_voltage = expected_voltage + tolerance_voltage

            self.assertGreaterEqual(
                supply_voltage,
                expected_minimum_voltage,
                (
                    f"STH supply voltage of {supply_voltage:.3f~} is lower "
                    "than expected minimum voltage of "
                    f"{expected_minimum_voltage:.3f~}"
                ),
            )
            self.assertLessEqual(
                supply_voltage,
                expected_maximum_voltage,
                (
                    f"STH supply voltage of {supply_voltage:.3f~} is "
                    "greater than expected maximum voltage of "
                    f"{expected_minimum_voltage:.3f~}"
                ),
            )

        self.loop.run_until_complete(test_supply_voltage())

    def test_acceleration_single_value(self):
        """Test stationary acceleration value"""

        async def test_acceleration_single():
            """Test stationary x acceleration value"""

            sensor = settings.acceleration_sensor()
            stream_data = await self.can.read_streaming_data_single()
            acceleration = convert_raw_to_g(
                stream_data.first.pop().value, sensor.acceleration.maximum
            )

            # We expect a stationary acceleration between -g₀ and g₀
            # (g₀ = 9.807 m/s²)
            expected_acceleration = g0(0)
            tolerance_acceleration = g0(sensor.acceleration.tolerance)
            expected_minimum_acceleration = (
                expected_acceleration - tolerance_acceleration
            )
            expected_maximum_acceleration = (
                expected_acceleration + tolerance_acceleration
            )

            self.assertGreaterEqual(
                acceleration,
                expected_minimum_acceleration,
                (
                    f"Measured acceleration {acceleration:.3f~} is lower "
                    "than expected minimum acceleration "
                    f"{expected_minimum_acceleration:.3f~}"
                ),
            )
            self.assertLessEqual(
                acceleration,
                expected_maximum_acceleration,
                (
                    f"Measured acceleration {acceleration:.3f~} is greater "
                    "than expected maximum acceleration "
                    f"{expected_maximum_acceleration:.3f~}"
                ),
            )

        self.loop.run_until_complete(test_acceleration_single())

    def test_acceleration_noise(self):
        """Test ratio of noise to maximal possible measurement value"""

        acceleration = [
            timestamped_value.value
            for timestamped_value in self.loop.run_until_complete(
                self.can.read_streaming_data_seconds(seconds=4)
            ).first
        ]

        cls = type(self)
        cls.ratio_noise_max = ratio_noise_max(acceleration)

        sensor = settings.acceleration_sensor()
        maximum_ratio_allowed = sensor.acceleration.ratio_noise_to_max_value
        self.assertLessEqual(
            cls.ratio_noise_max,
            maximum_ratio_allowed,
            (
                "The ratio noise to possible maximum measured value of "
                f"{cls.ratio_noise_max} dB is higher than the maximum allowed "
                f"level of {maximum_ratio_allowed} dB"
            ),
        )

    def test_acceleration_self_test(self):
        """Execute self test of accelerometer"""

        async def read_voltages(dimension, reference_voltage) -> List[int]:
            """Read acceleration voltages in millivolts"""

            before = await self.can.read_acceleration_voltage(
                dimension, reference_voltage
            )

            await self.can.activate_acceleration_self_test(dimension)
            between = await self.can.read_acceleration_voltage(
                dimension, reference_voltage
            )

            await self.can.deactivate_acceleration_self_test(dimension)
            after = await self.can.read_acceleration_voltage(
                dimension, reference_voltage
            )

            return [round(value * 1000) for value in (before, between, after)]

        sensor = settings.acceleration_sensor()

        (
            voltage_before_test,
            voltage_at_test,
            voltage_after_test,
        ) = self.loop.run_until_complete(
            read_voltages(sensor.self_test.dimension, sensor.reference_voltage)
        )

        voltage_diff = voltage_at_test - voltage_before_test
        voltage_diff_abs = abs(voltage_diff)
        voltage_diff_before_after = abs(
            voltage_before_test - voltage_after_test
        )

        # - Voltage difference can be both positive or negative
        # - Voltage before and after the self test should be roughly the same

        voltage_diff_expected = sensor.self_test.voltage.difference
        voltage_diff_tolerance = sensor.self_test.voltage.tolerance

        voltage_diff_minimum = voltage_diff_expected - voltage_diff_tolerance
        voltage_diff_maximum = voltage_diff_expected + voltage_diff_tolerance

        self.assertLessEqual(
            voltage_diff_before_after,
            voltage_diff_tolerance,
            (
                "Measured voltage difference between voltage before and after "
                f"test {voltage_diff_before_after:.0f} mV is larger than "
                f"tolerance of {voltage_diff_tolerance:.0f} mV"
            ),
        )

        possible_failure_reason = (
            "\n\nPossible Reason:\n\n• Acceleration sensor config value "
            f"“{settings.sth.acceleration_sensor.sensor}” is incorrect"
        )

        self.assertGreaterEqual(
            voltage_diff_abs,
            voltage_diff_minimum,
            (
                f"Measured voltage difference of {voltage_diff_abs:.0f} mV is "
                "lower than expected minimum voltage difference of "
                f"{voltage_diff_minimum:.0f} mV{possible_failure_reason}"
            ),
        )
        self.assertLessEqual(
            voltage_diff_abs,
            voltage_diff_maximum,
            (
                f"Measured voltage difference of {voltage_diff_abs:.0f} mV is "
                "greater than expected minimum voltage difference of "
                f"{voltage_diff_maximum:.0f} mV{possible_failure_reason}"
            ),
        )

    def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        async def test_eeprom():
            """Test the EERPOM of the STH"""

            cls = type(self)
            receiver = "STH 1"

            # ========
            # = Name =
            # ========

            name = (
                settings.sth.serial_number
                if settings.sth.status == "Epoxied"
                else convert_mac_base64(cls.bluetooth_mac)
            )

            await self.can.write_eeprom_name(name, receiver)
            read_name = await self.can.read_eeprom_name(receiver)

            self.assertEqual(
                name,
                read_name,
                (
                    f"Written name “{name}” does not match read name"
                    f" “{read_name}”"
                ),
            )

            cls.name = read_name

            # =========================
            # = Sleep & Advertisement =
            # =========================

            await self._test_eeprom_sleep_advertisement_times()

            # ================
            # = Product Data =
            # ================

            await self._test_eeprom_product_data(Node(receiver), settings.sth)

            # ==============
            # = Statistics =
            # ==============

            await self._test_eeprom_statistics(
                Node(receiver),
                settings.sth.production_date,
                settings.sth.batch_number,
            )

            # ================
            # = Acceleration =
            # ================

            sensor = settings.acceleration_sensor()
            acceleration_max = sensor.acceleration.maximum
            adc_max = 0xFFFF
            acceleration_slope = acceleration_max / adc_max
            acceleration_offset = -(acceleration_max / 2)

            async def write_read_check(
                class_variable, write_routine, value, read_routine, axis, name
            ):
                await write_routine(value)
                setattr(cls, class_variable, await read_routine())
                read_value = getattr(cls, class_variable)
                self.assertAlmostEqual(
                    value,
                    read_value,
                    msg=(
                        f"Written {axis} acceleration {name} "
                        f"“{acceleration_slope:.5f}” does not match read "
                        f"{axis} acceleration {name} "
                        f"“{read_value:.5f}”"
                    ),
                )

            class_variables = (
                "acceleration_slope_x",
                "acceleration_slope_y",
                "acceleration_slope_z",
                "acceleration_offset_x",
                "acceleration_offset_y",
                "acceleration_offset_z",
            )
            write_routines = (
                self.can.write_eeprom_x_axis_acceleration_slope,
                self.can.write_eeprom_y_axis_acceleration_slope,
                self.can.write_eeprom_z_axis_acceleration_slope,
                self.can.write_eeprom_x_axis_acceleration_offset,
                self.can.write_eeprom_y_axis_acceleration_offset,
                self.can.write_eeprom_z_axis_acceleration_offset,
            )
            read_routines = (
                self.can.read_eeprom_x_axis_acceleration_slope,
                self.can.read_eeprom_y_axis_acceleration_slope,
                self.can.read_eeprom_z_axis_acceleration_slope,
                self.can.read_eeprom_x_axis_acceleration_offset,
                self.can.read_eeprom_y_axis_acceleration_offset,
                self.can.read_eeprom_z_axis_acceleration_offset,
            )
            names = chain(*(repeat("slope", 3), repeat("offset", 3)))
            values = chain(
                *(
                    repeat(acceleration_slope, 3),
                    repeat(acceleration_offset, 3),
                )
            )
            axes = list("xyz") * 2

            for (
                class_variable,
                write_routine,
                value,
                read_routine,
                axis,
                name,
            ) in zip(
                class_variables,
                write_routines,
                values,
                read_routines,
                axes,
                names,
            ):
                await write_read_check(
                    class_variable,
                    write_routine,
                    value,
                    read_routine,
                    axis,
                    name,
                )

            # =================
            # = EEPROM Status =
            # =================

            await self._test_eeprom_status(Node(receiver))

            # =========
            # = Reset =
            # =========

            # We reset the STH and STU to make sure
            # - the name change takes place and we can connect to the STH
            #   using the new name
            # - the STH also takes the other changed EEPROM values (such as
            #   the changed advertisement times) into account.
            await self.can.reset_node("STH 1")
            await self.can.reset_node("STU 1")

            try:
                await self.can.connect_sensor_device(
                    cls.name
                )  # Reconnect to STH
            except TimeoutError:
                self.fail(
                    "Unable to reconnect to STH using updated name "
                    f"“{cls.name}”"
                )

        self.loop.run_until_complete(test_eeprom())


def main():
    unittest_main(
        testRunner=ExtendedTestRunner, module="mytoolit.test.production.sth"
    )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    main()
